#-------------------------------------------------------------------------------
# Settings
#-------------------------------------------------------------------------------

# Important: Make sure the settings below match the values you configured in the 
# bash-script 01_run_computations.sh.

SKIP_SUPPLEMENT <- 0 # set to 1 to skip replicating results from supplementary material
SKIP_PAPER <- 0 # set to 1 to skip replicating results from paper


#-------------------------------------------------------------------------------
# Load/install R libraries
#-------------------------------------------------------------------------------

options(repos="https://cloud.r-project.org")
pkgTest <- function(x,y="") {
  if (!require(x,character.only = TRUE)) {
    install.packages(x,dep=TRUE)
    if(!require(x,character.only = TRUE)) stop("Package not found")
  }
  return("OK")
}

global.libraries <- c("tidyverse","kableExtra","latex2exp","cowplot","gridExtra")
results <- sapply(as.list(global.libraries), pkgTest)


#-------------------------------------------------------------------------------
# Configure Plots and Tables
#-------------------------------------------------------------------------------

distributions <- c('N(0,1)', 't(15)','t(10)','t(5)','SKU','KU','BM','SPB','SKB','TRI')
colour_scales <- c('#000000', '#e41a1c', '#285b85', '#4daf4a', '#984ea2','#ff7f00')
linetype_scales <- c('solid','longdash', 'twodash', 'dotted', 'dotdash')
CS_colours <- c('#9ECAE1', '#3182BD')

theme_ng <- function() {
  theme_cowplot() + 
    theme(rect = element_rect(fill = "transparent", colour = NA, color = NA,
                              linewidth = 0, linetype = 0),
          legend.spacing = unit(12/4, "pt"),
          legend.key = element_blank(),
          legend.key.size = unit(1/3 * 12, "pt"),
          legend.key.width = unit(1.5 * 12, 'pt'),
          legend.text = element_text(size = rel(0.8)),
          legend.title = element_text(size = rel(0.8)),
          legend.position = 'bottom',
          legend.box = 'horizontal',
          legend.margin=margin(0,0,0,0),
          legend.box.margin=margin(-7,10,10,10),
          strip.background = element_rect(fill = "NA", colour = "black", linewidth = 0.125),
          plot.title = element_text(size=11, face="plain"),
          axis.title.x = element_text(size=11, face="plain"),
          axis.title.y = element_text(size=11, face="plain"),
          axis.text.x = element_text(size=9),
          axis.text.y = element_text(size=9),
          panel.spacing.y = unit(2, "mm"),
          panel.grid.major = element_line(colour = "#DDDDDD", size=rel(0.5)),
          panel.grid.minor = element_line(colour = "#EEEEEE", size=rel(0.25))
    )
}


#-------------------------------------------------------------------------------
# Replicate Paper
#-------------------------------------------------------------------------------

Table2 <- function() {
  raw <- read_csv('output/Table2.csv', show_col_types = FALSE) %>% arrange(id)
  tab <- raw %>% 
    pivot_longer(cols=c(score_ols, score_onestep), names_to='test', values_to="size") %>%
    mutate(size=size*100) %>%
    pivot_wider(id_cols=c(test,K,p,n), names_from=eta, values_from=size) %>%
    arrange(desc(test),K,p,n) %>% select(everything(),-test)
  
  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep=c(rep('', 2), "\\midrule")) %>% 
    row_spec(row=0, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{One-Step Efficient Estimates}}\\\\') %>%
    row_spec(row=18, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{OLS Estimates}}\\\\') %>%
    write('output/Table2.tex')
}


Table3 <- function() {
  stat_names <- c(
    'score_ols', 'score_onestep',
    'dw_gmm', 'dw_bkr',
    'pml_t_lm', 'pml_gm_lm', 'gmm_ll_lm', 'gmm_kew_lm',
    'pml_t_lr', 'pml_gm_lr', 'gmm_ll_lr', 'gmm_kew_lr',
    'pml_t_w', 'pml_gm_w', 'gmm_ll_w', 'gmm_kew_w'
  )
  stat_tex <- c(
    '$\\hat{S}_{ols}$', '$\\hat{S}_{onestep}$',
    '$\\rm S^{\\rm DW}$', '$\\rm BKR^{\\rm DW}$',
    '$\\rm LM^{\\rm PML,t}$', '$\\rm LM^{\\rm PML,gm}$', '$\\rm LM^{\\rm GMM,LL}$', '$\\rm LM^{\\rm GMM,Kew}$',
    '$\\rm LR^{\\rm PML,t}$', '$\\rm LR^{\\rm PML,gm}$', '$\\rm LR^{\\rm GMM,LL}$', '$\\rm LR^{\\rm GMM,Kew}$',
    '$\\rm W^{\\rm PML,t}$', '$\\rm W^{\\rm PML,gm}$', '$\\rm W^{\\rm GMM,LL}$', '$\\rm W^{\\rm GMM,Kew}$')
  
  # create table data frame
  tab <- read_csv('output/Table3.csv', show_col_types = FALSE) %>%
    mutate( eta=factor(eta, levels=distributions) ) %>%
    filter(!grepl('pml_gm_', statistic)) %>%
    mutate( statistic=factor(statistic, levels=stat_names, labels=stat_tex )) %>% 
    mutate( rejection_rate=rejection_rate*100 ) %>%
    arrange(desc(iid), eta, statistic) %>%
    pivot_wider(id_cols=c(iid,statistic), names_from=eta, values_from=rejection_rate) %>%
    rename(Test=statistic) %>%
    select(everything(), -iid)
  
  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep='') %>% 
    row_spec(row=0, extra_latex_after='\\midrule\\multicolumn{11}{l}{\\textit{$\\epsilon_{1,t} \\sim \\epsilon_{2,t}$}}\\\\') %>%
    row_spec(row=4, extra_latex_after='\\midrule') %>%
    row_spec(row=7, extra_latex_after='\\midrule') %>%
    row_spec(row=10, extra_latex_after='\\midrule') %>%
    write('output/Table3.tex')
}


Figure1 <- function() {
  tests <- c('score_ols', 'score_onestep', 'pml_t_lm', 'dw_gmm', 'dw_bkr')
  tests_label <- TeX( c('$\\hat{S}_{ols}$', '$\\hat{S}_{onestep}$', '$LM^{PML,t}$', '$S^{DW}$', '$BKR^{DW}$') )
  
  # load data
  data <- read_csv('output/Figure1.csv', show_col_types = FALSE) %>%
    filter(eta!='N(0,1)') %>%
    mutate(
      eta = factor(eta, levels=distributions),
      statistic = factor(statistic, levels=tests)
    )
  
  # plot power curves
  ggplot(data, aes(x=alpha, y=rejection_rate, linetype=statistic, color=statistic)) +
    geom_line(linewidth=.75) +
    facet_wrap(vars(eta)) +
    scale_colour_manual(values = colour_scales, labels=unname(tests_label)) +
    scale_linetype_manual(values = linetype_scales, labels=unname(tests_label)) +
    ylab("") +
    xlab(TeX("$\\alpha$")) +
    theme_bw() + 
    theme_ng() +
    theme(legend.title= element_blank())
  
  fn <- 'output/Figure1.pdf'
  print(fn)
  ggsave(file=fn, units="cm", height = 16, width = 20)
}


Figure2 <- function() {
  CS <- read_csv("output/Figure2.csv") %>%
    filter(level %in% c(0.33, 0.05)) %>%
    group_by(alpha_d, alpha_s) %>%
    summarise(coverage=sum(covered)) %>% 
    mutate(coverage=as.factor(coverage))
  
  ggplot(CS, aes(x=alpha_s, y=alpha_d)) +
    geom_tile(aes(fill=coverage)) +
    scale_fill_manual(values=CS_colours) +
    scale_x_continuous(
      breaks=seq(from=0.0, to=3.0, by=.5),
      minor_breaks=seq(from=0.0, to=3.0, by=.1),
      limits=c(0,3), expand=c(0,0)
    ) +
    scale_y_continuous(
      breaks=seq(from=-3.0, to=0.0, by=.5),
      minor_breaks=seq(from=-3.0, to=0.0, by=.1),
      limits=c(-3,0), expand=c(0,0)) +
    xlab(TeX("$\\alpha^s$")) + ylab(TeX("$\\alpha^d$")) + 
    theme_ng() + theme(legend.position="none")
  
  ggsave(file='output/Figure2.pdf', units="cm", height=2/3*20, width=20)
}


Figure3 <- function() {
  yLimits <- list(c(-3, 3), c(-1, 4))
  
  irf_bands <- read_csv("output/Figure3.csv") %>%
    mutate(
      level = as.factor(level),
      variable = factor(variable, levels=c(1, 2), labels=c("Real wage", "Employment")),
      shock = factor(shock, levels=c(1, 2), labels=c("Demand shock", "Supply shock")),
    ) %>% group_by(variable, shock)
  
  plts <- irf_bands %>% group_map(function (df, key) {
    
    p <- ggplot(df, aes(x=h, fill=level)) +
      geom_ribbon(aes(ymin=lower, ymax=upper)) +
      geom_hline(yintercept=0, linetype="dotted", color="black") +
      scale_fill_manual(values=CS_colours) +
      scale_x_continuous(
        breaks=seq(from=0, to=24, by=2),
        minor_breaks=seq(from=0, to=24, by=1),
        limits=c(0,24), expand=c(0,0)
      ) +
      scale_y_continuous(
        breaks=seq(from=yLimits[[key$variable]][1], to=yLimits[[key$variable]][2], by=1),
        limits=yLimits[[key$variable]], expand=c(0,0)
      ) +
      ylab(paste(key$variable)) +
      xlab("") +
      ggtitle(paste(key$shock)) +
      theme_ng() + theme(legend.position="none")
    
    return(p)
  })
  
  plt <- grid.arrange(grobs=plts, padding=unit(0.1, "line"))
  ggsave(plt, file='output/Figure3.pdf', units="cm", height=0.8*20, width=20)
}


Figure4 <- function() {
  CS <- read_csv("output/Figure4.csv") %>%
    filter(level %in% c(0.33, 0.05)) %>%
    group_by(alpha_qx, alpha_qp, level) %>%
    summarise(covered=max(covered)) %>%
    group_by(alpha_qx, alpha_qp) %>%
    summarise(coverage=sum(covered)) %>% 
    mutate(coverage=as.factor(coverage)) %>% ungroup()
  
  ggplot(CS, aes(x=alpha_qp, y=alpha_qx)) +
    geom_tile(aes(fill=coverage)) +
    scale_fill_manual(values=CS_colours) +
    scale_x_continuous(
      breaks=seq(from=0.0, to=0.1, by=.01),
      minor_breaks=seq(from=0.0, to=0.1, by=.005),
      limits=c(0,0.1), expand=c(0,0)
    ) +
    scale_y_continuous(
      breaks=seq(from=0.0, to=0.25, by=.05),
      minor_breaks=seq(from=0.0, to=0.25, by=.01),
      limits=c(0,0.25), expand=c(0,0)) +
    xlab(TeX("$\\alpha_{qp}$")) + ylab(TeX("$\\alpha_{qx}$")) + 
    geom_segment(aes(x=0.0, xend=0.0258, y=0.0258, yend=0.0258), linetype="longdash") +
    geom_segment(aes(x=0.0258, xend=0.0258, y=0.0258, yend=0), linetype="longdash") +
    theme_ng() + theme(legend.position="none")
  
  ggsave(file='output/Figure4.pdf', units="cm", height=2/3*20, width=20)
}


Figure5 <- function() {
  yLimits <- list(c(-5, 5), c(-10, 21), c(-10, 17))
  
  irf_bands <- read_csv("output/Figure5.csv") %>%
    mutate(
      level = as.factor(level),
      variable = factor(variable, levels=c(1, 2, 3), labels=c("Oil production", "Real activity", "Real price of oil")),
      shock = factor(shock, levels=c(1, 2, 3), labels=c("Oil supply shock", "Aggr. demand shock", "Oil spec. demand shock")),
    ) %>% group_by(variable, shock)
  
  plts <- irf_bands %>% group_map(function (df, key) {
    
    p <- ggplot(df, aes(x=h, fill=level)) +
      geom_ribbon(aes(ymin=lower, ymax=upper)) +
      geom_hline(yintercept=0, linetype="dotted", color="black") +
      scale_fill_manual(values=CS_colours) +
      scale_x_continuous(
        breaks=seq(from=0, to=16, by=5),
        minor_breaks=seq(from=0, to=16, by=1),
        limits=c(0,16), expand=c(0,0)
      ) +
      scale_y_continuous(
        breaks=seq(from=yLimits[[key$variable]][1], to=yLimits[[key$variable]][2], by=5),
        limit=yLimits[[key$variable]], expand=c(0,0)
      ) +
      ylab(paste(key$variable)) +
      xlab("") +
      ggtitle(paste(key$shock)) +
      theme_ng() + theme(legend.position="none")
    
    return(p)
  })
  
  plt <- grid.arrange(grobs=plts, padding=unit(0.1, "line"))
  ggsave(plt, file='output/Figure5.pdf', units="cm", height=0.8*20, width=20)
}

# Produce .tex and .pdf files for the paper
if (SKIP_PAPER < 1){
Table2()
Table3()
Figure1()
Figure2()
Figure3()
Figure4()
Figure5()
}


#-------------------------------------------------------------------------------
# Replicate Supplementary Material
#-------------------------------------------------------------------------------

TableS1 <- function() {
  raw <- read_csv('output/TableS1.csv', show_col_types = FALSE) %>% arrange(id)
  tab <- raw %>% 
    pivot_longer(cols=c(score_ols, score_onestep), names_to='test', values_to="size") %>%
    mutate(size=size*100) %>%
    pivot_wider(id_cols=c(test,K,p,n), names_from=eta, values_from=size) %>%
    arrange(desc(test),K,p,n) %>% select(everything(),-test)
  
  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep=c(rep('', 2), "\\midrule")) %>% 
    row_spec(row=0, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{One-Step Efficient Estimates}}\\\\') %>%
    row_spec(row=18, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{OLS Estimates}}\\\\') %>%
    write('output/TableS1.tex')
}


TableS2 <- function() {
  stat_names <- c(
    'score_ols', 'score_onestep',
    'dw_gmm', 'dw_bkr',
    'pml_t_lm', 'pml_gm_lm', 'gmm_ll_lm', 'gmm_kew_lm',
    'pml_t_lr', 'pml_gm_lr', 'gmm_ll_lr', 'gmm_kew_lr',
    'pml_t_w', 'pml_gm_w', 'gmm_ll_w', 'gmm_kew_w'
  )
  stat_tex <- c(
    '$\\hat{S}_{ols}$', '$\\hat{S}_{onestep}$',
    '$\\rm S^{\\rm DW}$', '$\\rm BKR^{\\rm DW}$',
    '$\\rm LM^{\\rm PML,t}$', '$\\rm LM^{\\rm PML,gm}$', '$\\rm LM^{\\rm GMM,LL}$', '$\\rm LM^{\\rm GMM,Kew}$',
    '$\\rm LR^{\\rm PML,t}$', '$\\rm LR^{\\rm PML,gm}$', '$\\rm LR^{\\rm GMM,LL}$', '$\\rm LR^{\\rm GMM,Kew}$',
    '$\\rm W^{\\rm PML,t}$', '$\\rm W^{\\rm PML,gm}$', '$\\rm W^{\\rm GMM,LL}$', '$\\rm W^{\\rm GMM,Kew}$')
  
  # create table data frame
  tab <- read_csv('output/TableS2.csv', show_col_types = FALSE) %>%
    mutate( eta=factor(eta, levels=distributions) ) %>%
    filter(!grepl('pml_gm_', statistic)) %>%
    mutate( statistic=factor(statistic, levels=stat_names, labels=stat_tex )) %>% 
    mutate( rejection_rate=rejection_rate*100 ) %>%
    arrange(desc(iid), eta, statistic) %>%
    pivot_wider(id_cols=c(iid,statistic), names_from=eta, values_from=rejection_rate) %>%
    rename(Test=statistic) %>%
    select(everything(), -iid)
  
  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep='') %>% 
    row_spec(row=0, extra_latex_after='\\midrule\\multicolumn{11}{l}{\\textit{$\\epsilon_{1,t} \\sim \\epsilon_{2,t}$}}\\\\') %>%
    row_spec(row=4, extra_latex_after='\\midrule') %>%
    row_spec(row=7, extra_latex_after='\\midrule') %>%
    row_spec(row=10, extra_latex_after='\\midrule') %>%
    write('output/TableS2.tex')
}


TableS3 <- function() {
  raw <- read_csv('output/TableS3.csv', show_col_types = FALSE) %>% arrange(id)
  tab <- raw %>% 
    pivot_longer(cols=c(score), names_to='test', values_to="size") %>%
    mutate(size=size*100) %>%
    pivot_wider(id_cols=c(test,K,p,n), names_from=eta, values_from=size) %>%
    arrange(desc(test),K,p,n) %>% select(everything(),-test)
  
  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep=c(rep('', 2), "\\midrule")) %>%
    write('output/TableS3.tex')
}


TableS4 <- function() {

  raw <- read_csv('output/TableS4.csv', show_col_types = FALSE) %>% arrange(id)
  tab <- raw %>%
    pivot_longer(cols=c(score_ols, score_onestep), names_to='test', values_to="size") %>%
    mutate(size=size*100) %>%
    pivot_wider(id_cols=c(test,K,p,n), names_from=eta, values_from=size) %>%
    arrange(desc(test),K,p,n) %>% select(everything(),-test)

  # write to .tex
  tab %>% kable(format='latex', digits=1, booktabs=T, escape=F, linesep=c(rep('', 1), "\\midrule")) %>%
    row_spec(row=0, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{One-Step Efficient Estimates}}\\\\') %>%
    row_spec(row=6, extra_latex_after='\\midrule\\multicolumn{12}{l}{\\textit{OLS Estimates}}\\\\') %>%
    write('output/TableS4.tex')
}


FigureS1_S2 <- function() {
  colour_scales <- c('#000000', '#e41a1c', '#4daf4a', '#984ea2','#ff7f00')
  linetype_scales <- c('solid','longdash', 'dotted', 'dotdash')
  
  tests <- c('score_ols', 'score_onestep', 'dw_gmm', 'dw_bkr')
  tests_label <- TeX( c('$\\hat{S}_{ols}$', '$\\hat{S}_{onestep}$', '$GMM^{DW}$', '$BKR^{DW}$') )
  
  data <- read_csv('output/FigureS1_S2.csv', show_col_types = FALSE) %>%
    filter(h>=0) %>%
    mutate(irf=as.factor((i-1)*max(j)+j)) %>%
    filter(irf==3) %>%
    filter(eta!='N(0,1)') %>% 
    group_by(n,p,K,eta,h,i,j,irf) %>% 
    pivot_longer(
      cols = c(coverage_score_ols,coverage_score_onestep,coverage_dw_gmm,coverage_dw_bkr,length_score_ols,length_score_onestep,length_dw_gmm,length_dw_bkr),
      names_to = 'measures',
      values_to = 'value'
    ) %>%
    separate(measures, into=c('measure','testname','testtype'), sep='_', fill='right') %>%
    unite(test, testname, testtype, sep='_', na.rm=T) %>%
    mutate(
      eta = factor(eta, levels=distributions),
      test = factor(test, levels=tests)
    )
  
  # Figure S1: Coverage
  ggplot(data %>% filter(measure=='coverage'), aes(x=h, y=value, linetype=test, color=test)) +
    geom_line(linewidth=.75) +
    geom_hline(yintercept=.9, linetype='dashed', color='gray') +
    facet_wrap(vars(eta)) +
    scale_x_continuous(breaks=seq(0,max(data$h),by=2)) +
    scale_colour_manual(values = colour_scales, labels=unname(tests_label)) +
    scale_linetype_manual(values = linetype_scales, labels=unname(tests_label)) +
    ylim(c(.85,1)) +
    ylab("Coverage") +
    xlab(TeX("Horizon ($H$)")) +
    theme_ng() +
    theme(legend.title= element_blank())
  ggsave(file='output/FigureS1.pdf', units = "cm", height = 16, width = 20)
  
  # Figure S2: Average length
  ggplot(data %>% filter(measure=='length'), aes(x=h, y=value, linetype=test, color=test)) +
    geom_line(linewidth=.75) +
    facet_wrap(vars(eta)) +
    scale_x_continuous(breaks=seq(0,max(data$h),by=2)) +
    scale_colour_manual(values = colour_scales, labels=unname(tests_label)) +
    scale_linetype_manual(values = linetype_scales, labels=unname(tests_label)) +
    ylab("Avg. Length") +
    xlab(TeX("Horizon ($H$)")) +
    theme_ng() +
    theme(legend.title= element_blank())
  
  ggsave(file='output/FigureS2.pdf', units = "cm", height = 16, width = 20)
  
}


TableS5 <- function() {
  method_names <- c('pml_t', 'gmm_ll', 'gmm_kew')
  
  # create table data frame
  tab <- read_csv('output/TableS5.csv', show_col_types = FALSE) %>% 
    mutate( eta=factor(eta, levels=distributions) ) %>%
    mutate( method=factor(method, levels=method_names)) %>% 
    arrange(eta, method) %>% 
    pivot_wider(id_cols=c(eta), names_from=c(method, type), values_from=rmse)
  
  # write to .tex
  tab %>% kable(format='latex', digits=3, booktabs=T, escape=F, linesep='', col.names=c('$\\eta$', rep(c('$\\hat{\\alpha}_n$','$\\hat{\\alpha}^{onestep}_n$'),3))) %>%
    add_header_above(c(" "=1, "$PML^{t}$"=2, "$GMM^{LL}$"=2, "$GMM^{Kew}$"=2), escape=F) %>%
    write('output/TableS5.tex')
}


FigureS3 <- function() {
  CS <- read_csv("output/FigureS3.csv") %>%
    group_by(alpha_d, alpha_s) %>%
    summarise(coverage=sum(covered)) %>% 
    mutate(coverage=as.factor(coverage))
  
  ggplot(CS, aes(x=alpha_s, y=alpha_d)) +
    geom_tile(aes(fill=coverage)) +
    scale_fill_manual(values=CS_colours) +
    scale_x_continuous(
      breaks=seq(from=0.0, to=3.0, by=.5),
      minor_breaks=seq(from=0.0, to=3.0, by=.1),
      limits=c(0,3), expand=c(0,0)
    ) +
    scale_y_continuous(
      breaks=seq(from=-3.0, to=0.0, by=.5),
      minor_breaks=seq(from=-3.0, to=0.0, by=.1),
      limits=c(-3,0), expand=c(0,0)) +
    xlab(TeX("$\\alpha^s$")) + ylab(TeX("$\\alpha^d$")) + 
    theme_ng() + theme(legend.position="none")
  
  ggsave(file='output/FigureS3.pdf', units="cm", height=2/3*20, width=20)
}


FigureS4 <- function() {
  yLimits <- list(c(-3, 3), c(-1.5, 4))
  
  irf_bands <- read_csv("output/FigureS4.csv") %>%
    mutate(
      level = as.factor(level),
      variable = factor(variable, levels=c(1, 2), labels=c("Real wage", "Employment")),
      shock = factor(shock, levels=c(1, 2), labels=c("Demand shock", "Supply shock")),
    ) %>% group_by(variable, shock)
  
  plts <- irf_bands %>% group_map(function (df, key) {
    
    p <- ggplot(df, aes(x=h, fill=level)) +
      geom_ribbon(aes(ymin=lower, ymax=upper)) +
      geom_hline(yintercept=0, linetype="dotted", color="black") +
      scale_fill_manual(values=CS_colours) +
      scale_x_continuous(
        breaks=seq(from=0, to=24, by=2),
        minor_breaks=seq(from=0, to=24, by=1),
        limits=c(0,24), expand=c(0,0)
      ) +
      scale_y_continuous(
        breaks=seq(from=yLimits[[key$variable]][1], to=yLimits[[key$variable]][2], by=1),
        limits=yLimits[[key$variable]], expand=c(0,0)
      ) +
      ylab(paste(key$variable)) +
      xlab("") +
      ggtitle(paste(key$shock)) +
      theme_ng() + theme(legend.position="none")
    
    return(p)
  })
  
  plt <- grid.arrange(grobs=plts, padding=unit(0.1, "line"))
  ggsave(plt, file='output/FigureS4.pdf', units="cm", height=0.8*20, width=20)
}


FigureS5 <- function() {
  data <- read_csv('output/FigureS5.csv', show_col_types = FALSE) %>%
    pivot_longer(c('eps1','eps2'))
  
  varnames <- c("Real wage", "Employment")
  
  i <<- 1
  plots <- data %>% group_by(name) %>% group_map(function (group, key){
    p <- ggplot(group, aes(x=value)) + 
      geom_histogram(aes(y=..density..), colour='gray', fill='white') + 
      geom_density(colour='black', linewidth=1.5) +
      stat_function(fun=dnorm, args=list(mean=0, sd=1), colour='red',linewidth=1, linetype='dashed') +
      xlab(TeX(sprintf('$\\epsilon_{%d,t}$',i))) + ylab("") +
      xlim(c(-5,5)) + ylim(c(0,0.5)) +
      ggtitle(varnames[i]) +
      theme_ng()
    i <<- i + 1
    return(p)
  })
  plt <- grid.arrange(grobs=plots, ncol=2)
  ggsave('output/FigureS5.pdf', plt, units = "cm", height = 10, width = 20)
}


FigureS6 <- function() {
  data <- read_csv('output/FigureS6.csv', show_col_types = FALSE) %>%
    pivot_longer(c('eps1','eps2','eps3'))
  
  varnames <- c("Oil production", "Real activity", "Real price of oil")
  
  i <<- 1
  plots <- data %>% group_by(name) %>% group_map(function (group, key){
    p <- ggplot(group, aes(x=value)) + 
      geom_histogram(aes(y=..density..), colour='gray', fill='white') + 
      geom_density(colour='black', linewidth=1.5) +
      stat_function(fun=dnorm, args=list(mean=0, sd=1), colour='red',linewidth=1, linetype='dashed') +
      xlab(TeX(sprintf('$\\epsilon_{%d,t}$',i))) + ylab("") +
      xlim(c(-5,5)) + ylim(c(0,0.6)) +
      ggtitle(varnames[i]) +
      theme_ng()
    i <<- i + 1
    return(p)
  })
  plt <- grid.arrange(grobs=plots, ncol=3)
  ggsave('output/FigureS6.pdf', plt, units = "cm", height=10, width = 30)
}

# Produce .tex and .pdf files for the Supplementary Material
if (SKIP_SUPPLEMENT < 1){
  TableS1()
  TableS2()
  TableS3()
  TableS4()
  FigureS1_S2()
  TableS5()
  FigureS3()
  FigureS4()
  FigureS5()
  FigureS6()
}
