#---
# Drautzburg & Wright (2021) GMM & BKR tests
#---

using LinearAlgebra, StatsFuns, SpecialFunctions, ForwardDiff

# Compute CS_α from grid
function compute_CS_α_DW(αs_grid::αGrid, γs::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, R::Int64)
    if !issorted(γs, rev=true); error("Error in get_CS_α: γs must be sorted in decreasing order."); end; # ensure γs are in decreasing order
    n_γs = length(γs)
    n_αs = length(αs_grid)
    
    # loop over grid points (αs)
    error_bool = falses(n_αs)
    CS_α_bool = falses(n_γs*2, n_αs)
    for i in 1:length(αs_grid)

        α_i = αs_grid[i]

        GMM_stat, BKR_stat = ComputeDwTests(α_i, Y, X, icsvar)

        for (j, γ) in pairs(γs)
            GMM_cv, BKR_cv = ComputeDwCritvals(α_i, Y, X, icsvar, R, γ)
            CS_α_bool[(j-1)*2+1,i] = (GMM_stat ≤ GMM_cv)
            CS_α_bool[(j-1)*2+2,i] = (BKR_stat ≤ BKR_cv)
        end
        error_bool[i] = false
        
    end

    CS_α_ind = [findall(CS_α_bool[i,:]) for i in axes(CS_α_bool,1)]
    error_ind = findall(error_bool)
    return CS_α_ind, error_ind
end

# GMM and BKR statistics
function ComputeDwTests(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR)

    β_hat, _ = estimate_β(α, Y, X, icsvar, false)
    Z_hat, _ = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)

    if icsvar.K!=2; error("Error in DW_stats. Only K=2 supported, since GMM moments hard-coded."); end

    A = get_A(α, β_hat[1:icsvar.d_β₁], icsvar)
    ϵ = A * Z_hat # K x n

    # BKR statistic
    BKR_stat = 0
    for i=1:icsvar.n
        BKR_stat += ( F_hat(ϵ, ϵ[:,i]) - prod( F_hat(ϵ, ϵ[:,i], true) ) )^2
    end

    # GMM statistic

    # define vector of moments
    n_s = binomial(icsvar.K+2,3) - icsvar.K
    n_k = binomial(icsvar.K+3,4) - icsvar.K
    g = Matrix(undef, n_s+n_k, icsvar.n)

    # skewness moments
    g[1,:] = ϵ[1,:].^2 .* ϵ[2,:]
    g[2,:] = ϵ[2,:].^2 .* ϵ[1,:]

    # kurtosis moments
    g[3,:] = ϵ[1,:].^2 .* ϵ[2,:].^2 .- 1
    g[4,:] = ϵ[1,:].^3 .* ϵ[2,:]
    g[5,:] = ϵ[2,:].^3 .* ϵ[1,:]

    gbar = mean(g, dims=2)

    # compute test statistic
    V = ((g .- gbar) * (g .- gbar)') / icsvar.n
    z = mean(g,dims=2)
    GMM_stat = (icsvar.n * (z' * (V \ z)))[1,1] # test statistic scalar

    GMM_stat, BKR_stat

end

# Bootstrap procedure for critical values
function ComputeDwCritvals(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, R::Int64, γ::Float64)

    β_hat, _ = estimate_β(α, Y, X, icsvar, false)
    Z_hat, _ = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)

    A = get_A(α, β_hat[1:icsvar.d_β₁], icsvar)
    Ainv = get_Ainv(α, β_hat[1:icsvar.d_β₁], icsvar)
    
    # get structural residuals
    ϵ_hat = A * Z_hat # K x icsvar.n

    # prepare matrices
    e_boot = Matrix{Float64}(undef, icsvar.K, icsvar.n + icsvar.p)
    Y_boot = Matrix{Float64}(undef, icsvar.K, icsvar.n + icsvar.p)
    Y_bs = view(Y_boot, :, icsvar.p+1:icsvar.n+icsvar.p)
    stats = Matrix{Float64}(undef, R, 2) # GMM & BKR stats

    for r in 1:R # bootstrap loop
        
        # obtain SVAR bootstrap sample
        bootstrap_sample!(Y_boot, e_boot, ϵ_hat, Y, Ainv, β_hat[icsvar.d_β₁+1:icsvar.d_β], icsvar.n, icsvar.p, icsvar.constant)

        X_bs = X_matrix(Y_boot, icsvar.p, icsvar.constant)

        # calculate DW statistic
        stats[r,1], stats[r,2] = ComputeDwTests(α, Y_bs, X_bs, icsvar)

    end

    quantile(stats[:,1],1-γ), quantile(stats[:,2],1-γ) # return critical values
    
end

# X is a (K x T) matrix, x a (K x 1) vector
function F_hat(X::AbstractMatrix, x::AbstractVector, marginals::Bool=false)
    I = (X .<= x)
    n = size(X, 2)
    if marginals
        sum(I,dims=2) / n
    else
        sum(all(I, dims=1)) / n
    end
end
