#---
# Density score approximation using B-Splines
#---

using StaticArrays, Statistics

# default settings for B-spline estimator
Base.@kwdef struct BSplines
    κ::Int64 = 4
    B::Int64 = 6
    adj::Real = 1.0
    cutoff::Real = 0.05
    crossValidate::Bool = false
end

# main function to compute density scores
function ComputeDensityScores(ϵ::AbstractMatrix, splines::BSplines)
    if splines.crossValidate
        return broadcast(k -> estimate_ϕ_cv(ϵ[:, k], splines.κ, splines.B, splines.adj, splines.cutoff), 1:size(ϵ, 2))
    else
        return broadcast(k -> estimate_ϕ(ϵ[:, k], splines.κ, splines.B, splines.adj, splines.cutoff), 1:size(ϵ, 2))
    end
end

# estimate log-density scores
function estimate_ϕ(X::AbstractVector, κ::Int64, B::Int64, adj::Real, cutoff::Real)
    n = length(X)
    K = B + κ
    endpoints = get_endpoints(X, adj, cutoff, n)
    ξ = collect(LinRange(endpoints[1], endpoints[2], K))
    b₋₁ = B_splines(X, ξ, κ-1)
    b = B_splines(X, ξ, b₋₁)
    c = B_spline_derivs(ξ, κ, b₋₁, n, K)
    γ = pinv(b' * b) * sum(c, dims = 1)'
    -b * γ
end

# estimate log-density scores with cross-validation
function estimate_ϕ_cv(X::AbstractVector, κ::Int64, Bmax::Int64, adj::Real, cutoff::Real)
    n = length(X)

    # select B via cross validation
    cvCrit = zeros(Bmax)
    for B in 2:Bmax
        cvCrit[B] = compute_cv_score(X, κ, B, adj, cutoff)
    end
    B_cv = findfirst((cvCrit[2:end-1] - cvCrit[3:end]) .< 0) + 1

    # estimate density scores
    return estimate_ϕ(X, κ, B_cv, adj, cutoff)
end

# compute cross-validation score
function compute_cv_score(X::AbstractVector, κ::Int64, B::Int64, adj::Real, cutoff::Real)
    n = length(X)
    mid = Int(floor(n/2))

    K = B + κ
    endpoints = get_endpoints(X, adj, cutoff, n)
    ξ = collect(LinRange(endpoints[1], endpoints[2], K))

    # compute γ for first half
    b₋₁1 = B_splines(X[1:mid], ξ, κ-1)
    b1 = B_splines(X[1:mid], ξ, b₋₁1)
    c1 = B_spline_derivs(ξ, κ, b₋₁1, mid, K)
    γ₁ = pinv(b1' * b1) * sum(c1, dims = 1)'

    # compute γ for second half
    b₋₁2 = B_splines(X[mid+1:end], ξ, κ-1)
    b2 = B_splines(X[mid+1:end], ξ, b₋₁2)
    c2 = B_spline_derivs(ξ, κ, b₋₁2, n-mid, K)
    γ₂ = pinv(b2' * b2) * sum(c2, dims = 1)'

    # compute cross-validation score
    c1g2 = γ₁' * ( 1/mid * (b2' * b2)) * γ₁ - 2 * γ₁' * mean(c2, dims = 1)'
    c2g1 = γ₂' * ( 1/(n-mid) * (b1' * b1)) * γ₂ - 2 * γ₂' * mean(c1, dims = 1)'
    crit = 0.5 * (c1g2 + c2g1)

    return crit[1]
end

# get B-spline endpoints
function get_endpoints(X::AbstractVector, adj::Real, cutoff::Real, n::Int64)
    Y = sort(X)
    q_adj  = adj * sqrt(log(log(n)))
    (max(Y[1], quantile!(Y, cutoff; sorted=true, alpha=1.0, beta=1.0) - q_adj), min(quantile!(Y, 1.0 - cutoff; sorted = true, alpha=1.0, beta=1.0) + q_adj, Y[end]))
end

# non-recursive: generate B-spline matrix of order κ = K - l + 1
function B_splines(X::AbstractVector{T}, ξ::AbstractVector, κ::Int64) where T
    n = length(X)
    K = length(ξ)
    
    # first order
    bk = Matrix{T}(undef, n, K-1)
    @inbounds @views for m = 1:K-1, i = 1:n
        if ξ[m] <= X[i] < ξ[m+1]
            bk[i, m] = one(T)
        else
            bk[i, m] = zero(T)
        end
    end
    if κ > 1
        @inbounds @views for k=2:κ
            bold = bk
            bk = Matrix{T}(undef, n, K-k)
            Δ = ξ[k:end] - ξ[1:end-k+1]
            for m = 1:K-k, i = 1:n
                bk[i, m] = ((X[i]-ξ[m])/ Δ[m]) * bold[i, m] + ((ξ[m+k] - X[i]) / Δ[m+1]) * bold[i, m+1]
            end           
        end
    end
    bk
end

# recursive: generate B-spline matrix of order κ = K - l + 1, given B-spline matrix of order κ-1, B₋₁.
function B_splines(X::AbstractVector{T},  ξ::AbstractVector, B₋₁::AbstractMatrix) where T
    n, l = size(B₋₁)
    K = length(ξ)
    k = K - l + 1
    B = Matrix{T}(undef, n, K-k)
    Δ = ξ[k:end] - ξ[1:end-k+1]
    @inbounds @views for m = 1:K-k, i = 1:n
        B[i, m] = ((X[i]-ξ[m])/ Δ[m]) * B₋₁[i, m] + ((ξ[m+k] - X[i]) / Δ[m+1]) * B₋₁[i, m+1]
    end 
    B
end

# generate B-spline derivatives
function B_spline_derivs(ξ::AbstractVector, κ::Int64, b1::AbstractMatrix{T}, n::Int64, K::Int64) where T
    c = Matrix{T}(undef, n, K - κ)
    d = κ - 1.0
    Δ = ξ[κ:end] - ξ[1:end-κ+1]
    @inbounds @views for m = 1:K-κ, i = 1:n
        c[i, m] = d * ((b1[i, m] / Δ[m]) - (b1[i, m+1] / Δ[m+1]))
    end
    c
end
