#---
# GMM Estimators and Tests for Non-Gaussian SVARs
# - Lanne & Luoto (2021, JBES)
# - Keweloh (2021, JBES)
#---

using LinearAlgebra, StatsFuns, SpecialFunctions, ForwardDiff, NLopt

# Main interface for testing H0: α=α₀ vs. HA: α≠α₀
function GMM_tests(α_H0::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64=1, init_vals::AbstractVector=[], bounds::AbstractVector=[[],[]], verbose::Bool=false, HAC_trunc = Int(round(icsvar.n^(1/4))))
    # moments = 1 => Lanne & Luoto (2021, JBES), one co-kurtosis condition eps1^3 * eps2
    # moments = 2 => Lanne & Luoto (2021, JBES), one co-kurtosis condition eps1 * eps2^3
    # moments = 3 => Keweloh (2021, JBES), both co-kurtosis conditions
    
    # unrestricted GMM estimator [H_A]
    Q_ur, θ_ur, _, _ = GMM_estimator([], Y, X, icsvar, moments, init_vals, bounds, verbose, HAC_trunc)

    # restricted GMM estimator [H0: α = α_0]
    Q_r, θ_r, _, d_m = GMM_estimator(α_H0, Y, X, icsvar, moments, init_vals, bounds, verbose, HAC_trunc)

    if verbose
        @printf "GMM Test Statistics\n"
    end
    
    # (1) LM statistic
    S_r = S_HAC(α_H0, θ_r[1:icsvar.d_β₁], θ_r[icsvar.d_β₁+1:end], Y, X, icsvar, moments, HAC_trunc)
    G = similar(Y, d_m, icsvar.d_α + icsvar.d_β)
    ForwardDiff.jacobian!(G, ξ -> GMM_mom(ξ[1:icsvar.d_α], ξ[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], ξ[icsvar.d_α+icsvar.d_β₁+1:end], Y, X, icsvar, moments, true), vcat(α_H0, θ_r))
    gbar = GMM_mom(α_H0, θ_r[1:icsvar.d_β₁], θ_r[icsvar.d_β₁+1:end], Y, X, icsvar, moments, true)
    BInv = (G' * (S_r \ G)) \ I(icsvar.d_α+icsvar.d_β)
    ∇_θQ = G' * (S_r \ gbar)
    LM = icsvar.n * ∇_θQ ⋅ ( BInv * ∇_θQ )
    if verbose; @printf("... LM statistic\t%g\n", LM); end
    
    # (2) Wald statistic
    R = vcat(I(icsvar.d_α), zeros(icsvar.d_β, icsvar.d_α))
    S_ur = S_HAC(θ_ur[1:icsvar.d_α], θ_ur[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], θ_ur[icsvar.d_α+icsvar.d_β₁+1:end], Y, X, icsvar, moments, HAC_trunc)
    δ = R' * θ_ur .- α_H0
    G = similar(Y, d_m, icsvar.d_α+icsvar.d_β)
    ForwardDiff.jacobian!(G, ξ -> GMM_mom(ξ[1:icsvar.d_α], ξ[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], ξ[icsvar.d_α+icsvar.d_β₁+1:end], Y, X, icsvar, moments, true), θ_ur)
    BInv = (G' * (S_ur \ G)) \ I(icsvar.d_α+icsvar.d_β)
    W = icsvar.n * δ' * ((R' * BInv * R) \ δ)
    if verbose; @printf("... Wald statistic\t%g\n", W); end
    
    # (3) LR statistic
    LR = - icsvar.n * (Q_ur - Q_r) # eq. (12) of Lanne & Luoto (2021, JBES)
    if verbose; @printf("... LR statistic\t%g\n\n", LR); end

    return LM, W, LR
end

function GMM_estimator(α_H0::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64=1, init_vals::AbstractVector=[], bounds::AbstractVector=[[],[]], verbose::Bool=false, HAC_trunc::Int64=Int(round(icsvar.n^(1/4))))
    # moments = 1 => Lanne & Luoto (2021, JBES), one co-kurtosis condition eps1^3 * eps2n
    # moments = 2 => Lanne & Luoto (2021, JBES), one co-kurtosis condition eps1 * eps2^3
    # moments = 3 => Keweloh (2021, JBES), both co-kurtosis conditions

    ## Process inputs
    if verbose
        if !isempty(α_H0)
            @printf "Restricted GMM Estimator\n"
        else
            @printf "Unrestricted GMM Estimator\n"
        end
        @printf("... Moment conditions\t%d\n", moments)
        @printf("... HAC truncation lags\t%d\n", HAC_trunc);
    end

    # set bounds for optimization procedure
    local θ_lb, θ_ub
    if length(bounds)!=2; error("Optimization bounds must be supplied as [[θ_lb], [θ_ub]]. Not supplied bounds will automatically be padded with -Inf/Inf."); end
    if length(bounds[1])==icsvar.d_α+icsvar.d_β && length(bounds[2])==icsvar.d_α+icsvar.d_β
        θ_lb = bounds[1]
        θ_ub = bounds[2]
    elseif length(bounds[1])==icsvar.d_α && length(bounds[2])==icsvar.d_α
        θ_lb = vcat(bounds[1], fill(-Inf, icsvar.d_β))
        θ_ub = vcat(bounds[2], fill(Inf, icsvar.d_β))
    else
        error("Supplied bounds have wrong length")
    end
    if verbose;
        @printf("... Bounds\t\tθ ∈ [[%s], [%s]]\n", join([string(p) for p in θ_lb], ", "), join([string(p) for p in θ_ub], ", "));
    end
    
    # set initial values
    local θ_init
    if isempty(init_vals);
        β_init, _ = estimate_β(α_H0, Y, X, icsvar, false) # OLS initial value
        θ_init = vcat(α_H0, β_init)
    else
        if length(init_vals)==icsvar.d_α
            β_init, _ = estimate_β(init_vals[1:icsvar.d_α], Y, X, icsvar, false) # OLS initial value
            θ_init = vcat(init_vals, β_init)
        elseif length(init_vals)==(icsvar.d_α+icsvar.d_β)
            θ_init = init_vals
        else
            error("Supplied initial values have wrong length.")
        end
    end
    if verbose;
        @printf("... Initialization\tθ = %s\n", join([string(p) for p in θ_init], ", "));
    end

    # check moment conditions
    if moments<1 || moments>3; error("Invalid moment condition type: Pick 1 (eps1^3*eps2), 2 (eps1*eps2^3) or 3 (both)."); end
    
    ## GMM Estimator
    L = size(X, 1)
    d_m = nr_moments(icsvar, L, moments)
    
    local Q_hat, θ_hat, ret_val_fs, ret_val
    if isempty(α_H0) # Unrestricted GMM estimator
        # Options
        opt_ur = NLopt.Opt(:LD_LBFGS, icsvar.d_α + icsvar.d_β)
        opt_ur.lower_bounds = θ_lb
        opt_ur.upper_bounds = θ_ub

        # First stage
        W = I(d_m)
        opt_ur.min_objective = (ξ, grad) -> GMM_obj_alt(ξ, grad, W, Y, X, icsvar, moments)
        (_, θ_ur_fs, ret_val_fs) = NLopt.optimize(opt_ur, θ_init)
        if verbose & (NLopt_success(ret_val_fs)<1)
            @printf("\nUnrestricted GMM estimator (first stage) optimization returned %s.", String(ret_val_fs))
        end

        # Second stage
        S_ur_fs = S_HAC(θ_ur_fs[1:icsvar.d_α], θ_ur_fs[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], θ_ur_fs[icsvar.d_α+icsvar.d_β₁+1:end], Y, X, icsvar, moments, HAC_trunc)
        W_ur_eff = S_ur_fs \ I(d_m)
        opt_ur.min_objective = (ξ, grad) -> GMM_obj_alt(ξ, grad,  W_ur_eff, Y, X, icsvar, moments)
        (Q_hat, θ_hat, ret_val) = NLopt.optimize(opt_ur, θ_ur_fs)
        if verbose & (NLopt_success(ret_val)<1)
            @printf("\nUnrestricted GMM estimator (second stage) optimization returned %s.", String(ret_val))
        end
    else
        # Options
        opt_r = NLopt.Opt(:LD_LBFGS, icsvar.d_β)
        opt_r.lower_bounds = θ_lb[icsvar.d_α+1:end]
        opt_r.upper_bounds = θ_ub[icsvar.d_α+1:end]
        
        # First stage
        W = I(d_m)
        opt_r.min_objective = (ξ, grad) -> GMM_obj_null(ξ, grad, α_H0, W, Y, X, icsvar, moments)
        (_, β_hat_fs, ret_val_fs) = NLopt.optimize(opt_r, θ_init[icsvar.d_α+1:end])
        if verbose & (NLopt_success(ret_val_fs)<1)
            @printf("\nRestricted GMM estimator (first stage) optimization returned %s.", String(ret_val_fs))
        end
        
        # Second stage
        S_r_fs = S_HAC(α_H0, β_hat_fs[1:icsvar.d_β₁], β_hat_fs[icsvar.d_β₁+1:end], Y, X, icsvar, moments, HAC_trunc)
        W_r_eff = S_r_fs \ I(d_m)
        opt_r.min_objective = (ξ, grad) -> GMM_obj_null(ξ, grad, α_H0, W_r_eff, Y, X, icsvar, moments)
        (Q_hat, θ_hat, ret_val) = NLopt.optimize(opt_r, β_hat_fs)
        if verbose & (NLopt_success(ret_val)<1)
            @printf("\nRestricted GMM estimator (second stage) optimization returned %s.", String(ret_val))
        end
    end
    
    if verbose;
        @printf("... Result W_init\t%s\n", String(ret_val_fs));
        @printf("... Result W_eff\t%s\n", String(ret_val));
        @printf("... Estimate\t\tθ = %s\n", join([string(p) for p in θ_hat], ", "));
        @printf("... GMM Objective val.\tS = %g\n", Q_hat);
        @printf("\n")
    end
    return Q_hat, θ_hat, ret_val, d_m
end


#
# Helper functions
#

# GMM optimization objective function under null hypothesis H_0
function GMM_obj_null(ξ::AbstractVector, grad::AbstractVector, α::AbstractVector, W::AbstractMatrix, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64)
    if length(grad) > 0
        ForwardDiff.gradient!(grad, ξ -> GMM_obj(α, ξ[1:icsvar.d_β₁], ξ[icsvar.d_β₁+1:end], W, Y, X, icsvar, moments), ξ)
    end
    return GMM_obj(α, ξ[1:icsvar.d_β₁], ξ[icsvar.d_β₁+1:end], W, Y, X, icsvar, moments)
end

# GMM optimization objective function under alternative hypothesis H_A
function GMM_obj_alt(ξ::AbstractVector, grad::AbstractVector, W::AbstractMatrix, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64)
    if length(grad) > 0
        ForwardDiff.gradient!(grad, ξ -> GMM_obj(ξ[1:icsvar.d_α], ξ[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], ξ[icsvar.d_α+icsvar.d_β₁+1:end], W, Y, X, icsvar, moments), ξ)
    end
    return GMM_obj(ξ[1:icsvar.d_α], ξ[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], ξ[icsvar.d_α+icsvar.d_β₁+1:end], W, Y, X, icsvar, moments)
end

# optimization objective function
function GMM_obj(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, W::AbstractMatrix, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64)
    gbar = GMM_mom(α, β₁, β₂, Y, X, icsvar, moments, true)
    return gbar ⋅ (W * gbar)
end

# Moment conditions
function GMM_mom(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64, returnMean::Bool)

    # recover estimates for structural shocks
    Z, _ = SVAR_residuals(β₂, Y, X)
    ϵ = get_A(α, β₁, icsvar) * Z

    # determine number of moments
    L = size(X, 1)
    d_g = nr_moments(icsvar, L, moments)

    # Lanne & Luoto (2021, JBES) Moment conditions
    g = similar(β₁, d_g, icsvar.n)
    @inbounds @views for t=1:icsvar.n
        m=0
        for k=1:icsvar.K, l=1:L
            g[m+=1, t] = ϵ[k, t] * X[l, t] # (5a): icsvar.K * L SVAR conditions
        end
        # K^2 variance conditions
        for k=1:icsvar.K
            g[m+=1, t] = ϵ[k, t]^2 - 1.0 # (5b): icsvar.K variance conditions
        end
        for i=1:icsvar.K, j=i+1:icsvar.K
            g[m+=1, t] = ϵ[i, t] * ϵ[j, t] # (5c) K(K-1)/2 orthogonality conditions
            g[m+=1, t] = ϵ[i, t]^2 * ϵ[j, t]^2 - 1.0 # K(K-1)/2 symmetric co-kurtosis conditions
            # asymmetric co-kurtosis conditions
            if moments==1 || moments==3
                g[m+=1, t] = ϵ[i, t]^3 * ϵ[j, t] # K(K-1)/2 asymmetric co-kurtosis conditions
            end
            if moments==2 || moments==3
                g[m+=1, t] = ϵ[i, t] * ϵ[j, t]^3 # # K(K-1)/2 asymmetric co-kurtosis conditions
            end
        end
    end
    
    # return moment vector
    if returnMean
        return mean(g, dims=2)
    else
        return g
    end
end

# HAC covariance estimator for covariance of moment conditions
function S_HAC(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, moments::Int64, trunc::Int64)
    g = GMM_mom(α, β₁, β₂, Y, X, icsvar, moments, false)
    S_HAC = g * g' / icsvar.n # Γ_0
    for j=1:trunc-1
        Γ_j = g[:,j+1:icsvar.n] * g[:,1:icsvar.n-j]' / icsvar.n
        ω_j = (1-abs(j/trunc)) # Bartlett kernel
        S_HAC += ω_j .* (Γ_j + Γ_j')
    end
    return S_HAC
end

# calculate number of moment conditions
function nr_moments(icsvar, L, moments)
    d_m = icsvar.K*(L + 1) + 3*Int(icsvar.K*(icsvar.K-1)/2)
    if moments==3; d_m += Int(icsvar.K*(icsvar.K-1)/2); end
    return d_m
end
