#---
# Helper functions
#---

using SparseArrays, LinearAlgebra

#
# 1 - Misc. Helper functions
#

function vech(A::AbstractMatrix{T}) where T
    m = LinearAlgebra.checksquare(A)
    v = Vector{T}(undef, (m*(m+1))>>1)
    k = 0
    @inbounds for j = 1:m, i = j:m
        v[k += 1] = A[i,j]
    end
    return v
end

function vec2ltri(v::AbstractVector{T}) where T
    n = length(v)
    s = round(Int,(sqrt(8n+1)-1)/2)
    s*(s+1)/2 == n || error("vec2ltri: length of vector is not triangular")
    k=0;
    Array{T}([ i>=j ? (k+=1; v[k]) : 0 for i=1:s, j=1:s ])
end

function duplication(m::Int64)
    i = collect(1:m^2)
    J = zeros(Int64, m, m)
    z = 1
    @inbounds for x = 1:m, y=x:m
        J[x,y] = J[y,x] = z  # duplication occurs here
        z += 1
    end
    # pass 3 vectors {i-index, j-index, value} to sparse()
    return sparse(i, vec(J), 1)
end

function elimination(m::Int64)
    vech_indices = vec(tril(trues(m, m)))
    I(m^2)[vech_indices, :]
end

function commutation(m::Int64, n::Int64)
    A = reshape(1:m*n, m, n)
    w = reshape(A', m*n)
    return I(m*n)[w, :]
end

function commutation(m::Int64)
    return commutation(m, m)
end

# modified sign function: returns 1 if positive/zero, 0 if negative
function sgn(x)
    Int(sign(x)>=0)*2-1
end

# mean of a vector, ignoring NaNs
function nanmean(x)
    mean(filter(!isnan, x))
end

# function to get indices from a grid of vectors
function get_α_from_grid(n, α_vecs...)
    α_inds = Vector{Int}(undef, length(α_vecs)) # indices 
    α_vals = Vector{Float64}(undef, length(α_vecs)) # values
    @inbounds for i in eachindex(α_vecs)
       l = length(α_vecs[i])
       ix = mod1(n, l)
       α_inds[i] = ix
       α_vals[i] = α_vecs[i][ix]
       n = cld(n, l)
    end
    (α_vals, α_inds) # return values & indices
end

function banded(n,k,v)
    return Array{Float64}([ abs(i-j)==k ? v : 0.0 for i=1:n, j=1:n ])
end

#
# 2 - SVAR Helper functions
#

# get companion matrix from VAR parameter vector
function B_companion(β₂::AbstractVector{T}, p::Int64, K::Int64, constant::Bool) where T
    if constant
        β₂ = β₂[K+1:end]
    end
    if p == 1
        reshape(β₂, K, K)
    else
        B = Matrix{T}(undef, K*p, K*p)
        B[K+1:end, 1:K*(p-1)] .= I(K*(p-1))
        B[K+1:end, K*(p-1)+1:end] .= 0.0
        @views @inbounds for (i, x) in enumerate(Iterators.partition(β₂, K^2))
            B[1:K, (i-1)*K+1: i*K] = reshape(x, K, K)
        end
        B
    end
end

# build matrix of constant and lags without padding
function X_matrix!(X::AbstractMatrix{T}, Y::AbstractMatrix, p::Int64, constant::Bool) where T
    K = size(Y,1)
    l = constant ? 1 : 0
    if constant
        X[1, :] .= one(T)
    end
    @views @inbounds for i=1:p
        X[l+1+(i-1)*K:l+i*K, :] = Y[:, p-i+1:end-i]
    end
end

# convenience interface without pre-allocation
function X_matrix(Y::AbstractMatrix{T}, p::Int64, constant::Bool) where T
    K, n = size(Y)
    n -= p
    X = constant ? Matrix{T}(undef, K*p+1, n) : Matrix{T}(undef, K*p, n)
    X_matrix!(X, Y, p, constant)
    X
end

# function to compute SVAR residuals
function SVAR_residuals(β₂::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix)
    B = reshape(β₂, (size(Y,1), size(X, 1)))
    Z = Y - B * X
    Σ_Z = 1/size(Y,2) * Z * Z'
    return (Z, Σ_Z)
end

# eq-by-eq OLS for Y a Kxn matrix
function OLS(Y::AbstractMatrix, X::AbstractMatrix)
    ((X * X') \ (X * Y'))'
end


#
# Grid functions
#

abstract type AbstractαGrid end

# default grid type - just a collection of vectors
mutable struct αGrid <: AbstractαGrid
    αs::AbstractVector
    n::Int64
    ns::Vector{Int64}
    vecs::AbstractVector
end

# implement indexing & length
Base.getindex(G::αGrid, i::Int64) = G.αs[i]
Base.length(G::αGrid) = G.n

# alternative constructor to build grid from list of vectors by taking all combinations
function create_α_grid(α_vecs...)
    αs = vec([collect(x) for x in Iterators.product(α_vecs...)])
    ns = [length(v) for v in α_vecs]
    αGrid(αs, length(αs), ns, collect(α_vecs))
end

# NLopt return codes
function NLopt_success(code::Symbol)
    if code==NLopt.SUCCESS || code==NLopt.STOPVAL_REACHED || code==NLopt.FTOL_REACHED || code==NLopt.XTOL_REACHED
        return 1
    elseif code==NLopt.MAXEVAL_REACHED || code==NLopt.MAXTIME_REACHED
        return 0
    else
        return -1
    end
end
