#---
# Functions to compute IRFs and confidence bands
#---

using Random, Distributions, LinearAlgebra, StaticArrays, StatsFuns, SpecialFunctions, ProgressMeter, Printf, ForwardDiff

# Default settings for IRFs (bands)
Base.@kwdef struct IrfSettings
    K::Int64
    H::Int64 # max. horizon for IRFs
    γs::Vector{Real} # list of CS levels
    γ₁s::Vector{Real} = γs ./ 2 # α test CI level
    γ₂s::Vector{Real} = γs .- γ₁s 
    cumulative::Vector{Bool} = falses(K^2) # default: no cumulative IRFs
end

#
# Algorithm 1
#

# Default: All αs admissible (no restrictions)
function α_admissible(α::AbstractVector, β_hat::AbstractVector, icsvar::AbstractICSVAR)
    return true
end

# Compute CS_α from grid
function compute_CS_α(αs_grid::αGrid, γs::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines, efficient::Bool, progress=false)

    if !issorted(γs, rev=true); error("Error in get_CS_α: γs must be sorted in decreasing order."); end; # ensure γs are in decreasing order
    n_γs = length(γs)
    n_αs = length(αs_grid)
    
    # Loop over grid points (αs)
    error_bool = falses(n_αs)
    CS_α_bool = falses(n_γs, n_αs)
    for i in 1:n_αs

        if progress!=false; update!(progress, i); end
        
        α_i = αs_grid[i]

        # estimate nuisance parameters
        local β_hat
        try
            if efficient
                β_hat, _ = estimate_β_onestep(α_i, Y, X, icsvar, splines, false)
            else
                β_hat, _ = estimate_β(α_i, Y, X, icsvar, false)
            end
        catch e
            error_bool[i] .= true
            continue
        end
        
        # drop α_i from CS if not admissible
        if !α_admissible(α_i, β_hat, icsvar)
            continue
        end

        # compute score test
        try
            α_i_accepted = test_αᵢ_in_CS(α_i, β_hat, γs, Y, X, icsvar, splines)
            CS_α_bool[:, i] = α_i_accepted
        catch e
            error_bool[i] = true
        end

    end

    # return CS_α
    CS_α_ind = [findall(CS_α_bool[i,:]) for i in axes(CS_α_bool,1)]
    error_ind = findall(error_bool)
    return CS_α_ind, error_ind
end

# helper function. returns boolean vector of whether α is included in CS for different levels
function test_αᵢ_in_CS(αᵢ::AbstractVector, β_hat::AbstractVector, γs::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)

    accept = falses(length(γs))

    # compute score statistic
    S_stat, S_rank = ComputeScoreTest(αᵢ, β_hat, Y, X, icsvar, splines)

    for (j, γ) in pairs(γs)
        if S_rank>0
            cv = quantile(Chisq(S_rank), 1-γ) # get critical value
            if (S_stat ≤ cv); accept[j] = true; end
        else
            error("Rank of efficient info matrix is 0.")
        end
    end
    
    return accept
end


#
# Algorithm 2 - Compute IRF confidence sets from CS_α
#

# main routine to compute IRF comfidence bands from CS_α obtained from Algorithm 1
function compute_IRF_bands(CS_α::Vector{Vector{Int64}}, αs_grid::αGrid, γ₂s::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, irfs::IrfSettings, splines::BSplines, efficient::Bool, progress=false)

    # perform checks on inputs
    if length(CS_α)!=length(γ₂s); error("Wrong confidence levels supplied."); end
    if !issorted(γ₂s, rev=true); error("Error in get_CS_α: γs must be sorted in decreasing order."); end; # ensure γs are in decreasing order
    
    # initialize results lists
    CS_irfs_lower = Matrix{Float64}[]
    CS_irfs_upper = Matrix{Float64}[]
    pr = 1;
    for (j, γ₂) in pairs(γ₂s) # loop through confidence levels
        
        lower = fill(NaN, icsvar.K^2*(irfs.H+1), length(CS_α[j]))
        upper = fill(NaN, icsvar.K^2*(irfs.H+1), length(CS_α[j]))
        l = 1;
        for i in CS_α[j] # loop through αs in CS_α
            
            αᵢ = αs_grid[i]

            # estimate β and Σ_β
            if efficient
                β_hat, Σᵦ_hat = estimate_β_onestep(αᵢ, Y, X, icsvar, splines, true)
            else
                β_hat, Σᵦ_hat = estimate_β(αᵢ, Y, X, icsvar, true)
            end

            # construct IRF CS for αᵢ and γ₂
            lower[:,l], upper[:,l] = CS_irfs_from_CS_α(αᵢ, β_hat, Σᵦ_hat, icsvar, γ₂, irfs)
            
            if progress!=false; update!(progress, pr); end
            l += 1; pr += 1;
        end
        
        push!(CS_irfs_lower, lower)
        push!(CS_irfs_upper, upper)
    end

    return CS_irfs_lower, CS_irfs_upper
end


#
# 2 - Auxiliary routines to compute IRFs & IRF CSs
#

# computes 1-γ₂ CS for IRFs via Δ-method with Bonferroni adjustment from CS(α)
function CS_irfs_from_CS_α(α::AbstractVector, β::AbstractVector, Σ_β::AbstractMatrix, icsvar::AbstractICSVAR, γ₂::Real, irfs::IrfSettings)

    # Bonferroni adjustment: Get critical value
    cv2 = [quantile(Normal(), 1-γ₂/2) for k = 1:icsvar.K^2]

    # for each horizon, construct a CI for the IRF
    νs = compute_irfs(α, β[1:icsvar.d_β₁], β[icsvar.d_β₁+1:icsvar.d_β], icsvar, irfs)
    ses = IRF_se_grid_num(α, β, Σ_β, icsvar, irfs)

    adj = repeat(vec(cv2), outer=(irfs.H+1,1)) .* ses # use same cv2 at every horizon
    lower = νs - adj
    upper = νs + adj

    return (lower, upper)
end

# Computes a K^2(H+1) vector of SVAR IRFs
# The IRFs are stacked by horizon with blocks of size K^2
# IRFs as defined in Ch. 4, Kilian & Lütkepohl 2012 eq. (4.1.1)ff
function compute_irfs(α::AbstractVector, β₁::AbstractVector{T1}, β₂::AbstractVector{T2}, icsvar::AbstractICSVAR, irfs::IrfSettings, h::Int64=-1) where {T1, T2}
    # type hack to establish ForwardDiff compatibility
    S=Float64; if T1!=Float64; S = T1; end; if T2!=Float64; S = T2; end

    J = [I(icsvar.K) zeros(icsvar.K, icsvar.K*(icsvar.p-1))]

    # initializations
    singleHorizon = (h>=0)
    nₕ = singleHorizon ? 1 : irfs.H+1
    hs = singleHorizon ? h : 0:irfs.H
    v = Vector{S}(undef, icsvar.K^2 * nₕ)
    C = Matrix{S}(undef, size(J,1), size(J,1))
    B = B_companion(β₂, icsvar.p, icsvar.K, icsvar.constant)

    # compute irfs for all horizons
    j = 0
    @views @inbounds for h in hs
        M = J * B^h * J'
        C = Matrix{S}(undef, size(J,1), size(J,1))
        mul!(C, M, get_Ainv(α, β₁, icsvar))

        v[j*icsvar.K^2+1:(j+1)*icsvar.K^2] = vec(C)
        j += 1
    end

    # transform to cumulative IRFs if necessary
    if any(irfs.cumulative)
        vR = reshape(v, (icsvar.K^2, irfs.H+1))
        vR[irfs.cumulative, :] = cumsum(vR[irfs.cumulative, :], dims=2)
        v = vec(vR)
    end
    return v
end


# computes IRF standard errors for a grid of horizons
function IRF_se_grid_num(α::AbstractVector, β::AbstractVector, Σ_β::AbstractMatrix, icsvar::AbstractICSVAR, irfs::IrfSettings)
    H = irfs.H
    se = Vector{Float64}(undef, icsvar.K^2*(H+1))

    # compute gradient of irfs w.r. to parameters for Δ-method
    ∇_β = ForwardDiff.jacobian((b) -> compute_irfs(α, b[1:icsvar.d_β₁], b[icsvar.d_β₁+1:icsvar.d_β], icsvar, irfs), β)

    # compute IRF se for every h in 0:H
    param_sel = [1:icsvar.d_β₁; icsvar.d_β₁+icsvar.constant*icsvar.K+1:icsvar.d_β] # selects all parameters except constant
    @views @inbounds for h in 0:H 
        F = ∇_β[h*icsvar.K^2+1:(h+1)*icsvar.K^2, param_sel] * Σ_β[param_sel, param_sel] * ∇_β[h*icsvar.K^2+1:(h+1)*icsvar.K^2, param_sel]'
        se[h*icsvar.K^2+1:(h+1)*icsvar.K^2] .= sqrt.(diag(F))
    end
    se  # return vectorized se matrix
end

# checks coverage for IRF union confidence sets
function check_coverage(lower::AbstractMatrix, upper::AbstractMatrix, irf₀::AbstractVector, K::Int64, H::Int64)
    flag = zeros(Bool, K^2*(H+1))
    @views @inbounds for i in axes(lower, 2)
        if all(flag)
            break
        end
        for l in eachindex(flag)
            if flag[l]
                continue
            end
            flag[l] = lower[l, i] ≤ irf₀[l] ≤ upper[l, i]
        end
    end
    flag
end
