#---
# Tests for independence of structural shocks ϵ_k by
# - (1) Montiel Olea, Plagborg-Moller and Qian (2022)
# - (2) Matteson & Tsay (2017), based on R package steadyICA implementation
#---

using LinearAlgebra, Printf

# Tests for independence of components ϵ_k [Montiel Olea, Plagborg-Moller and Qian, 2022]
function IndependenceTests(γ_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, R::Int64=1000)

    # Montiel Olea, Plagborg-Moller and Qian (2022)
    pval_mopmq = MOPMQ_pval(γ_hat, Y, X, icsvar, R)

    # Matteson & Tsay (2017)
    pval_mt = MT_pval(γ_hat, Y, X, icsvar, R)

    return pval_mopmq, pval_mt
end


#
# (1) Montiel Olea, Plagborg-Moller and Qian (2022)
#
function MOPMQ_pval(γ_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, R::Int64)

    α_hat = γ_hat[1:icsvar.d_α]
    β₁_hat = γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁]
    β₂_hat = γ_hat[icsvar.d_α+icsvar.d_β₁+1:end]

    # compute statistic
    stat = MOPMQ_statistic(γ_hat, Y, X, icsvar)

    Z_hat, _ = SVAR_residuals(β₂_hat, Y, X)
    A = get_A(α_hat, β₁_hat, icsvar)
    Ainv = get_Ainv(α_hat, β₁_hat, icsvar)
    ϵ_hat = A * Z_hat

    # prepare matrices
    e_boot = Matrix{Float64}(undef, icsvar.K, icsvar.n + icsvar.p)
    Y_boot = Matrix{Float64}(undef, icsvar.K, icsvar.n + icsvar.p)
    Y_bs = view(Y_boot,:, icsvar.p+1:icsvar.n+icsvar.p)
    stats = Vector{Float64}(undef,R)

    for r in 1:R # bootstrap loop

        # obtain SVAR bootstrap sample
        bootstrap_sample!(Y_boot, e_boot, ϵ_hat, Y, Ainv, β₂_hat, icsvar.n, icsvar.p, icsvar.constant)
        X_bs = X_matrix(Y_boot, icsvar.p, icsvar.constant)
        
        # calculate DW statistic
        stats[r] = MOPMQ_statistic(γ_hat, Y_bs, X_bs, icsvar)
        
    end
    
    return (1 + sum(stats .>= stat))/(R + 1) # bootstrapped p-value
end

function MOPMQ_statistic(γ_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR)
    α_hat = γ_hat[1:icsvar.d_α]
    β₁_hat = γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁]
    β₂_hat = γ_hat[icsvar.d_α+icsvar.d_β₁+1:end]

    Z_hat, _ = SVAR_residuals(β₂_hat, Y, X)
    A = get_A(α_hat, β₁_hat, icsvar)
    ϵ2_hat = (A * Z_hat).^2

    Shat = 0.0
    for i=1:icsvar.K, j=1:icsvar.K
        if i==j; continue; end
        Shat += ( (ϵ2_hat[i,:] ⋅ ϵ2_hat[j,:]) / icsvar.n )^2
    end
    return sqrt( Shat / (icsvar.n*(icsvar.n-1)) )
end


#
# (2) Matteson & Tsay (2017)
#

function MT_pval(γ_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, R::Int64)
    
    α_hat = γ_hat[1:icsvar.d_α]
    β₁_hat = γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁]
    β₂_hat = γ_hat[icsvar.d_α+icsvar.d_β₁+1:end]

    Z_hat, _ = SVAR_residuals(β₂_hat, Y, X)
    A = get_A(α_hat, β₁_hat, icsvar)
    ϵ_hat = A * Z_hat

    # permutation test
    S = ϵ_hat'
    p = 1.0
    n, d = size(S)
    obs = MT_gmultidcov(S, p)
    over = 0
    @inbounds @views for r=1:R
        S1 = zeros(n, d)
        for i=1:d
            S1[:, i] = S[sample(axes(S, 1), n), i]
        end
        stat = MT_gmultidcov(S1, p)
        if stat >= obs
            over += 1
        end
    end
    return (over + 1)/(R + 1)

end

# U statistic
function MT_dcovustat(X::AbstractMatrix, Y::AbstractMatrix, p::Real = 1.0)
    n_X, d_X = size(X)
    n_Y, d_Y = size(Y)
    if n_X != n_Y
        DimensionMismatch("X and Y must have the same number of rows")
    end

    T1 = T2x = T2y = T3 = 0.0
    Xsum = zeros(n_X)
    Ysum = zeros(n_Y)

    X = X / n_X
    Y = Y / n_Y

    @inbounds @views for k = 2:n_X 
        for l = 1:(k-1)
            # Calculate |X_k - X_l|
            dX = 0.0
            for i = 1:d_X
                dX += (X[k, i] - X[l, i])^2
            end
            dX = sqrt(dX)^p

            # Calculate |Y_k - Y_l|
            dY = 0.0
            for j=1:d_Y
                dY += (Y[k, j] - Y[l, j])^2
            end
            dY = sqrt(dY)^p

            # Update T1, k-th elem of Xsum, Ysum
            T1 += dX * dY
            Xsum[k] += dX
            Xsum[l] += dX
            Ysum[k] += dY
            Ysum[l] += dY
        end
        # Update T2x, T2y
        T2x += Xsum[k]
        T2y += Ysum[k]
    end
    @inbounds @views for k=1:n_X
        T3 += Xsum[k] * Ysum[k] / 3
    end
    
    # Update 
    T3 -=  2*(T1)/3
    T1 = 2 * T1 * (n_X / (n_X - 1))
    T2x = 2 * T2x / (n_X - 1)
    T2y = 2 * T2y / (n_X - 1)
    T3 = 6 * T3 * (n_X / ((n_X - 1) * (n_X - 2)))
    T2 = T2x * T2y
    
    # Return
    T1 + T2 - T3
end

function MT_gmultidcov(S::AbstractMatrix, p::Real=1.0)
    d = size(S, 2)
    out = 0.0
    grps = 1:d
    for i=grps
        out += MT_dcovustat(S[:, i:i], S[:, [j!=i for j in grps]], p)
    end
    out
end
