#---
# Baumeister & Hamilton (2015) ICSVAR Model
# α = (α_d, α_s)'
#---

using StaticArrays, LinearAlgebra

# define ICSVAR type
Base.@kwdef struct ICSVAR_BH15 <: AbstractICSVAR
    n::Int64
    p::Int64
    K::Int64
    d_α::Int64
    constant::Bool
    d_β₁::Int64
    # optional arguments
    d_β₂::Int64 = p*K^2 + constant*K
    d_β::Int64 = d_β₁ + d_β₂
    d_γ::Int64 = d_α + d_β
end

# load model
function load_model(type::String="direct")
    # load raw data
    data = DataFrame(CSV.File(string(Base.@__DIR__)*"/../raw_data/BH15-dataset.csv", header=0))
    rename!(data, [:date, :wage, :employment])
    wage = [100*(log(data[i,"wage"]) - log(data[i-1,"wage"])) for i ∈ 2:size(data,1)]
    employment = [100 *(log(data[i,"employment"]) - log(data[i-1,"employment"])) for i ∈ 2:size(data,1)]
    
    # build data objects
    p = 8
    constant = true
    YY = hcat(wage[92:end], employment[92:end])'
    Y = YY[:,p+1:end]
    X = X_matrix(YY, p, constant)

    # return model
    if type=="direct"
        return Y, X, ICSVAR_BH15(
            n=size(Y,2), p=p, K=2, d_α=2, constant=constant, d_β₁=2
        )
    elseif type=="euler"
        return Y, X, ICSVAR_canonical(
            n=size(Y,2), p=p, K=2, d_α=1, constant=constant, d_β₁=3 
        )
    else
        error("Unknown type.")
    end
end


#
# 1 - A^-1(α, σ) and A(α, σ)
#

# α = (α_d, α_s)'
function get_Ainv(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_BH15)
    if length(α)!=2 || length(β₁)!=2; error("Error in BH15 parameterization of Ainv: d_α!=2 or d_β₁!=2."); end
    c = α[2]-α[1]
    return SMatrix{2, 2}(β₁[1]/c, β₁[1]*α[2]/c, -β₁[2]/c, -β₁[2]*α[1]/c)
end

function get_A(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_BH15)
    if length(α)!=2 || length(β₁)!=2; error("Error in BH15 parameterization of Ainv: d_α!=2 or d_β₁!=2."); end
    return SMatrix{2, 2}(-α[1]/β₁[1], -α[2]/β₁[2], 1.0/β₁[1], 1.0/β₁[2])
end

function get_A_grad(α::AbstractVector, β₁::AbstractVector, l::Int64, icsvar::ICSVAR_BH15)
    if length(α)!=2 || length(β₁)!=2; error("Error in BH15 parameterization of ∇A: d_α!=2 or d_β₁!=2."); end
    if l<1 || l>length(α)+length(β₁); error("Error in BH15 parameterization of ∇A: Gradient of A only defined for 1<=l<=4."); end
    if l==1; return SMatrix{2, 2}(-1/β₁[1], 0.0, 0.0, 0.0); end # derivative w.r. to α_d
    if l==2; return SMatrix{2, 2}(0.0, -1/β₁[2], 0.0, 0.0); end # derivative w.r. to α_s
    if l==3; return SMatrix{2, 2}(α[1]/β₁[1]^2, 0.0, -1.0/β₁[1]^2, 0.0); end # derivative w.r. to β₁[1]
    if l==4; return SMatrix{2, 2}(0.0, α[2]/β₁[2]^2, 0.0, -1.0/β₁[2]^2); end # derivative w.r. to β₁[2]
end


#
# 2 - β estimation
#

# OLS for β₂ and Method of Moments for β₁
function estimate_β(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::ICSVAR_BH15, compute_Σᵦ::Bool=false)

    # 1 - Estimate β₂ by OLS
    B_hat = OLS(Y, X)
    β₂_hat = vec(B_hat)

    # 2 - Estimate β₁ by method of moments
    (_, Σ_Z) = SVAR_residuals(β₂_hat, Y, X)
    B0 = SMatrix{2, 2}(-α[1], -α[2], 1.0, 1.0)
    BΣB = B0 * Σ_Z * B0'
    β₁_hat = [sqrt(BΣB'[1,1]), sqrt(BΣB'[2,2])]

    β_hat = vcat(β₁_hat, β₂_hat)
    if !compute_Σᵦ;
        return β_hat, nothing
    end

    # 3 - Compute Σᵦ
    error("BH15: Σᵦ estimator not implemented.")
    #return (β_hat, Σᵦ_hat)
end


#
# 3 - Alternative parametrisation (Euler angles)
#

# compute elasticites from scalar (α, β₁)
function compute_elasticities(α::AbstractVector, β₁::AbstractVector)
    α_d = ( β₁[2]*sin(α[1]) - β₁[3]*cos(α[1]) ) / ( β₁[1]*sin(α[1]) )
    α_s = ( β₁[2]*cos(α[1]) + β₁[3]*sin(α[1]) ) / ( β₁[1]*cos(α[1]) )
    return vcat(α_d, α_s)
end

# computes 1-γ₂ CS for elasticities via Δ-method with Bonferroni adjustment from CS(α)
function compute_CS_elasticities(CS_α::Vector{Vector{Int64}}, αs_grid::αGrid, gs_grid::αGrid, γ₂s::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines, efficient::Bool, progress=false)
    
    # perform checks on inputs
    if length(CS_α)!=length(γ₂s); error("Wrong confidence levels supplied."); end
    if !issorted(γ₂s, rev=true); error("Error in get_CS_α: γs must be sorted in decreasing order."); end; # ensure γs are in decreasing order
    
    # initialize results lists
    n_γs = length(γ₂s)
    n_gs = length(gs_grid)
    CS_g_bool = falses(n_γs, n_gs)

    pr = 1
    for (j, γ₂) in pairs(γ₂s) # loop through confidence levels
    
        for i in CS_α[j] # loop through αs in CS_α

            α_i = αs_grid[i]

            # estimate β and Σ_β
            if efficient
                β_hat, Σᵦ_hat = estimate_β_onestep(α_i, Y, X, icsvar, splines, true)
            else
                β_hat, Σᵦ_hat = estimate_β(α_i, Y, X, icsvar, true)
            end
            
            # compute 
            g_hat = compute_elasticities(α_i, β_hat[1:icsvar.d_β₁])

            # compute gradient of (α_d, α_s) w.r. to β
            J_γ = ForwardDiff.jacobian((b) -> compute_elasticities(α_i, b), β_hat[1:icsvar.d_β₁])
            V_hat_inv = J_γ * Σᵦ_hat[1:icsvar.d_β₁,1:icsvar.d_β₁] * J_γ' \ I(size(J_γ,1))
            cv = quantile(Chisq(size(J_γ,1)), 1-γ₂)

            # loop through all possible values of g()
            for k in 1:length(gs_grid)
                
                # obtain confidence interval
                if ((g_hat - gs_grid[k])' * V_hat_inv * (g_hat - gs_grid[k]) ≤ cv)
                    CS_g_bool[j,k] = true # union of confidence sets
                end

            end

            if progress!=false; update!(progress, pr); end
            pr += 1;

        end

    end
    
    CS_g_ind = [findall(CS_g_bool[i,:]) for i in axes(CS_g_bool,1)]
    return CS_g_ind
end
