#---
# Canonical ICSVAR Model
#
# A^-1 = Σ_12(β₁) * R(α)
#
# where Σ_12(β₁) is a lower triangular matrix and R(α) is a rotation matrix which is parameterized by either
#
# (i) Euler angles (for K=2 and K=3)
# (ii) Caley transform (for any K)
#---

using StaticArrays, LinearAlgebra

# define default ICSVAR types
Base.@kwdef struct ICSVAR_canonical <: AbstractICSVAR
    n::Int64
    p::Int64
    K::Int64
    d_α::Int64
    constant::Bool
    d_β₁::Int64
    # optional arguments
    d_β₂::Int64 = p*K^2 + constant*K
    d_β::Int64 = d_β₁ + d_β₂
    d_γ::Int64 = d_α + d_β
    R_type::String = "trig" # trig or caley
end

# OLS + Method of moments estimator
# Note: α is only passed to keep the syntax general.
function estimate_β(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::ICSVAR_canonical, compute_Σᵦ::Bool=false)

    # 1 - Estimate β₂ by OLS
    B_hat = OLS(Y, X)
    β₂_hat = vec(B_hat)

    # 2 - Estimate β₁ by Cholesky factorization
    (_, Σ_Z) = SVAR_residuals(β₂_hat, Y, X)
    β₁_hat = vech(cholesky(Σ_Z).L)
    β_hat = vcat(β₁_hat, β₂_hat)

    if !compute_Σᵦ
        return β_hat, nothing
    end

    # 3 - Compute Σᵦ
    Σᵦ_hat = OLSChol_get_Σᵦ(β_hat, Y, X, icsvar)
    return β_hat, Σᵦ_hat
end

# Ainv(α, β₁) for trig parameterization
function get_Ainv(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_canonical)
    if icsvar.R_type=="caley"
        return get_Σ_sqrt_LT(β₁) * get_R_caley(α, icsvar.K)'
    elseif icsvar.R_type=="trig"
        return get_Σ_sqrt_LT(β₁) * get_R_orth_trig(α)
    else
        error("Unsupported value for icsvar.R_type.")
    end
end

# A(α,β₁) for trig parameterization
function get_A(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_canonical)
    if icsvar.R_type=="caley"
        return get_R_caley(α, icsvar.K) * inv(get_Σ_sqrt_LT(β₁))
    elseif icsvar.R_type=="trig"
        return get_R_orth_trig(α)' * inv(get_Σ_sqrt_LT(β₁))
    else
        error("Unsupported value for icsvar.R_type.")
    end
end

# ∇A(α, β₁) for trig parameterization
function get_A_grad(α::AbstractVector, β₁::AbstractVector, l::Int64, icsvar::ICSVAR_canonical)
    if icsvar.R_type=="caley"
        if 1<=l & l<=length(α) # derivative w.r. to α
            return (get_Σ_sqrt_LT(β₁)' \ get_R_caley_grad(α,l,icsvar.K)')'
        else
            return -get_R_caley(α, icsvar.K) * inv(get_Σ_sqrt_LT(β₁)) * get_Σ_sqrt_LT_grad(length(β₁), l-length(α)) * inv(get_Σ_sqrt_LT(β₁))
        end
    elseif icsvar.R_type=="trig"
        if 1<=l & l<=length(α) # derivative w.r. to α
            return (get_Σ_sqrt_LT(β₁)' \ get_R_orth_trig_grad(α,l))'
        else
            return -inv(get_R_orth_trig(α)) * inv(get_Σ_sqrt_LT(β₁)) * get_Σ_sqrt_LT_grad(length(β₁), l-length(α)) * inv(get_Σ_sqrt_LT(β₁))
        end
    end
end


#
# 1 - Helper functions for Σ_12(β₁)
#

function get_Σ_sqrt_LT(β₁::AbstractVector)
    vec2ltri(β₁)
end

function get_Σ_sqrt_LT_grad(dβ₁::Int64, l::Int64)
    v = zeros(dβ₁)
    v[l] = 1.0
    vec2ltri(v)
end

function get_Σ_sqrt_LT_grad(l::Int64)
    if l<1 || l>4; error("Invalid argument: l ∉ [1,2,3,4]."); end 
    D = SMatrix{2, 2}(0.0, 0.0, 0.0, 0.0)
    setindex(D, 1.0, l == 3 ? 4 : l)
end


#
# 2 - Helper functions for Euler parameterization of R(α)
#

# Trig. rotation matrix
function get_R_orth_trig(α::AbstractVector)
    if length(α)==1; return get_R_orth_trig2(α[1]); end
    if length(α)==3; return get_R_orth_trig3(α); end
    error("Error in get_R_orth_trig: Invalid length(α). Only K=2,3 supported.");
end

# Trig. rotation matrix gradient
function get_R_orth_trig_grad(α::AbstractVector, l::Int64)
    if length(α)==1; return get_R_orth_trig_grad2(α[1]); end
    if length(α)==3; return get_R_orth_trig_grad3(α, l); end
    error("Error in get_R_orth_trig_grad: Invalid length(α). Only K=2,3 supported.");
end

# Trig. rotation matrix, K=2
function get_R_orth_trig2(α::Real)
    c = cos(α)
    s = sin(α)
    SMatrix{2,2}(c, s, -s, c)
end

# Trig. rotation matrix gradient, K=2
function get_R_orth_trig_grad2(α::Real)
    c = cos(α)
    s = sin(α)
    SMatrix{2, 2}(-s, c, -c, -s)
end

# Trig. rotation matrix, K=3
function get_R_orth_trig3(α::AbstractVector)
    if length(α)!=3; error("Error in get_R_orth_trig3: length(α)!=3."); end
    (a, b, c) = α
    ca = cos(a)
    cb = cos(b)
    cc = cos(c)
    sa = sin(a)
    sb = sin(b)
    sc = sin(c)
    Rz = SMatrix{3,3}(ca,sa,0,-sa,ca,0,0,0,1)
    Ry = SMatrix{3,3}(cb,0,-sb,0,1,0,sb,0,cb)
    Rx = SMatrix{3,3}(1,0,0,0,cc,sc,0,-sc,cc)
    Rz * Ry * Rx
end

# Trig. rotation matrix gradient, K=3
function get_R_orth_trig_grad3(α::AbstractVector,l::Int64)
    if length(α)!=3; error("Error in get_R_orth_trig3: length(α)!=3."); end
    (a, b, c) = α
    ca = cos(a)
    cb = cos(b)
    cc = cos(c)
    sa = sin(a)
    sb = sin(b)
    sc = sin(c)
    Rz = SMatrix{3,3}(ca,sa,0,-sa,ca,0,0,0,1)
    Ry = SMatrix{3,3}(cb,0,-sb,0,1,0,sb,0,cb)
    Rx = SMatrix{3,3}(1,0,0,0,cc,sc,0,-sc,cc)
    if l==1
        Dz = SMatrix{3,3}(-sa,ca,0,-ca,-sa,0,0,0,0)
        Dz * Ry * Rx
    elseif l==2
        Dy = SMatrix{3,3}(-sb,0,-cb,0,0,0,cb,0,-sb)
        Rz * Dy * Rx
    elseif l==3
        Dx = SMatrix{3,3}(0,0,0,0,-sc,cc,0,-cc,-sc)
        Rz * Ry * Dx
    else
        error("Invalid argument: l ∉ [1,2,3].")
    end
end


#
# 3 - Helper functions for Caley parameterization of R(α)
#

# Rotation matrix from caley transform
function get_R_caley(α::AbstractVector{T}, K::Int64) where T
    # make skew-symmetric matrix O
    O = Matrix{T}(undef,K,K)
    fill!(O, 0.0)
    s = 1
    for k in 1:K
        O[k,k+1:end] = α[s:s+K-k-1]
        O[k+1:end,k] = -α[s:s+K-k-1]
        s = s + length(α[s:s+K-k-1])
    end
    # Caley transform
    R = (I(K) - O) * inv(I(K) + O)
end

# Rotation matrix from caley transform
function get_R_caley_grad(α::AbstractVector{T}, l::Int64, K::Int64) where T
    # make skew-symmetric matrix O
    O = Matrix{T}(undef, K, K)
    fill!(O, 0.0)
    s = 1
    for k in 1:K
        O[k,k+1:end] = α[s:s+K-k-1]
        O[k+1:end,k] = -α[s:s+K-k-1]
        s = s + length(α[s:s+K-k-1])
    end

    # make derivative of skew-symmetric matrix O
    L = Vector{T}(undef, length(α))
    fill!(L, 0.0)
    L[l] = 1
    Odl = zeros(K, K)
    s = 1
    for k in 1:K
       Odl[k,k+1:end] = L[s:s+K-k-1]
       Odl[k+1:end,k] = -L[s:s+K-k-1]
       s = s + length(L[s:s+K-k-1])
    end
 
    # make derivative of R
    Rderl = -Odl * inv(I(K) + O) - (I(K) - O) * inv(I(K) + O) * Odl * inv(I(K) + O)
end


#
# 4 - Helper function for parameter estimation
#

# required in asymp. formula below
function estimate_Qinv(X::AbstractMatrix)
    (X * X' / size(X, 2)) \ I
end

# required in asymp. formula below
function estimate_V₂₂(Z::AbstractMatrix{T}, return_VV::Bool=false) where T
    K, n = size(Z)
    v = Matrix{T}(undef, (K*(K+1))>>1, n)
    @views @inbounds for (i, z) in enumerate(eachcol(Z))
        v[:, i] = vech(z * z')
    end
    if return_VV
        (v, cov(v, dims = 2))
    else
        cov(v, dims = 2)
    end
end

function OLSChol_get_Σᵦ(β_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR)

    # compute residuals
    (Z, Σ_Z) = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)

    elimₖ = elimination(icsvar.K)
    commₖ = commutation(icsvar.K)
    N₂₂ = elimₖ * (I(icsvar.K^2) + commutation(icsvar.K))

    # compute Σ_β₁
    L = cholesky(Σ_Z).L
    V = estimate_V₂₂(Z)
    N = (N₂₂ * kron(L, I(icsvar.K)) * elimₖ') \ I
    Σ_β₁ = N * V * N' / icsvar.n
    
    # compute Σ_β₂
    Qinv = estimate_Qinv(X)
    Σ_β₂ = kron(Qinv, Σ_Z)
    Σ_β₂ = Σ_β₂ / icsvar.n

    # Stack Σ_β₁, Σ_β₂ diagonally
    Σ_β = zeros(icsvar.d_β, icsvar.d_β)
    Σ_β[1:icsvar.d_β₁, 1:icsvar.d_β₁] = Σ_β₁
    Σ_β[icsvar.d_β₁+1:icsvar.d_β, icsvar.d_β₁+1:icsvar.d_β] = Σ_β₂
    return Σ_β
end
