#---
# Kilian and Murphy (2012) Model (K=3)
#---

using DelimitedFiles, Statistics, ForwardDiff, StaticArrays, LinearAlgebra, NLopt, LaTeXStrings

abstract type ICSVAR_KM12 <: AbstractICSVAR end

Base.@kwdef struct ICSVAR_KM12_DIRECT <: ICSVAR_KM12
    n::Int64
    p::Int64
    K::Int64
    d_α::Int64
    constant::Bool
    d_β₁::Int64
    # optional arguments
    d_β₂::Int64 = p*K^2 + constant*K
    d_β::Int64 = d_β₁ + d_β₂
    d_γ::Int64 = d_α + d_β
end

function load_model()

    # load dataset
    data = readdlm(string(Base.@__DIR__)*"/../raw_data/Zhou20-dataset1.txt")
    data = data[:,1:3]
    data[:,1] = data[:,1] * 100
    data[:,3] = (data[:,3] .- mean(data[:,3])) * 100

    # build data objects
    p = 12
    constant = true
    YY = data'
    Y = YY[:,p+1:end]
    X = X_matrix(YY, p, constant) 

    return Y, X, ICSVAR_KM12_DIRECT(n = size(Y,2), p=p, K=3, d_α=3, constant=constant, d_β₁=6)
end

# α = (α_qx, α_qp, α_xp)'
function get_Ainv(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_KM12_DIRECT)
    if length(α)!=3 || length(β₁)!=6; error("Error in KM12 parameterization of Ainv: d_α!=3 or d_β₁!=6."); end
    SMatrix{3, 3}(β₁[1],β₁[2],β₁[3],α[1]*β₁[5],β₁[4],β₁[5],α[2]*β₁[6],α[3],β₁[6])
end

function get_A(α::AbstractVector, β₁::AbstractVector, icsvar::ICSVAR_KM12_DIRECT)
    if length(α)!=3 || length(β₁)!=6; error("Error in KM12 parameterization of Ainv: d_α!=3 or d_β₁!=6."); end
    get_Ainv(α, β₁, icsvar) \ I(3)
end

function get_A_grad(α::AbstractVector, β₁::AbstractVector, l::Int64, icsvar::ICSVAR_KM12_DIRECT)
    if length(α)!=3 || length(β₁)!=6; error("Error in KM12 parameterization of Ainv: d_α!=3 or d_β₁!=6."); end
    if l<1 || l>length(α)+length(β₁); error("Error in KM12 parameterization of ∇A: Gradient of A only defined for 1<=l<=9."); end
    if 1<=l & l<=length(α) # derivative w.r. to α
        ∇A_α = ForwardDiff.jacobian(a -> get_A(a, β₁, icsvar), α)
        return ∇A_α[(l-1)*3+1:(l*3),:]
    else # derivative w.r. to β₁
        ∇A_β₁ = ForwardDiff.jacobian(b -> get_A(α, b, icsvar), β₁)
        return reshape(∇A_β₁[:,l-3], (3,3))
    end
end

# OLS for β₂ and Method of Moments for β₁
function estimate_β(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::ICSVAR_KM12_DIRECT, compute_Σᵦ::Bool=false)

    # 1 - Estimate β₂ by OLS
    B_hat = OLS(Y, X)
    β₂_hat = vec(B_hat)

    # 2 - Estimate β₁ by method of moments estimation
    (_, Σ_Z) = SVAR_residuals(β₂_hat, Y, X)
    (β₁_hat, err) = KM12_β₁_MM(α, Σ_Z, icsvar)

    β_hat = vcat(β₁_hat, β₂_hat)
    if !compute_Σᵦ
        return β_hat, err;
    end

    # 3 - Compute Σᵦ
    error("KM12: Σᵦ estimator not implemented.")
    #return (β_hat, Σᵦ_hat)
end

# method-of-moments objective function for estimating β₁
function KM12_obj_mm(σ::AbstractVector, α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_KM12_DIRECT)
    Ainv = get_Ainv(α, σ, icsvar)
    m = vech(Ainv*Ainv') - vech(Σ) # 6x1 vector of moment conditions
    return m'*m
end

# NLopt objective function
function KM12_obj_opt(σ::AbstractVector, grad::AbstractVector, α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_KM12_DIRECT)
    if length(grad)>0
        ForwardDiff.gradient!(grad, s -> KM12_obj_mm(s, α, Σ, icsvar), σ)
    end
    return KM12_obj_mm(σ, α, Σ, icsvar)
end

# estimation routine for β₁
function KM12_β₁_MM(α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_KM12_DIRECT)

    β₁_init = [1.567, 0.007, -0.835, 4.242, 2.211, 5.526] # β_1 estimate from KM12
    
    # Options
    opt = NLopt.Opt(:LD_LBFGS, icsvar.d_β₁)
    opt.min_objective = (β₁, grad) -> KM12_obj_opt(β₁, grad, α, Σ, icsvar)
    (_, β₁_hat, ret_code) = NLopt.optimize(opt, β₁_init)

    # check for optimization failures
    if NLopt_success(ret_code)<1
        @printf("KM12: β₁ optimization returned with status %s for α=", String(ret_code))
        display(α')
        return (β₁_hat, true)
    else
        return (β₁_hat, false)
    end
end

# Overwrite one-step updates, to impose restrictions on β₁
function estimate_β_onestep(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::ICSVAR_KM12_DIRECT, splines::BSplines, compute_Σᵦ::Bool=false)

    # 1 - obtain initial estimate from function provided by model declaration
    β_hat, _ = estimate_β(α, Y, X, icsvar, false)
    
    # 2 - efficient score update
    β_hat = effscore_update_β(β_hat, α, Y, X, icsvar, splines)
    
    # project unrestricted estimates onto sign restriction set
    sr = [1.0, 1.0, -1.0, 1.0, 1.0, 1.0]
    violations = findall( sgn.(β_hat[1:icsvar.d_β₁] .* sr) .< 0 )
    β_hat[violations] .= sr[violations] .* 1e-9

    if !compute_Σᵦ; return β_hat, nothing; end
    
    # 3 - compute Σᵦ
    Σᵦ_hat = effscore_compute_Σᵦ(β_hat, α, Y, X, icsvar, splines)
    return β_hat, Σᵦ_hat
end

function α_admissible(α::AbstractVector, β_hat::AbstractVector, icsvar::ICSVAR_KM12_DIRECT)
    return true
end
