#---
# Triangular ICSVAR Model
#
# A^-1 = L + U
#
# where L is a lower triangular matrix (including diagonal) and
# U is an upper triangular matrix (excluding diagonal).
#---

using StaticArrays, LinearAlgebra, NLopt

# define default ICSVAR types
Base.@kwdef struct ICSVAR_triangular <: AbstractICSVAR
    n::Int64
    p::Int64
    K::Int64
    d_α::Int64
    constant::Bool
    d_β₁::Int64
    # optional arguments
    d_β₂::Int64 = p*K^2 + constant*K
    d_β::Int64 = d_β₁ + d_β₂
    d_γ::Int64 = d_α + d_β
end

# OLS + Method of moments estimator
function estimate_β(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::ICSVAR_triangular, compute_Σᵦ::Bool=false)

    # 1 - Estimate β₂ by OLS
    B_hat = OLS(Y, X)
    β₂_hat = vec(B_hat)

    # 2 - Estimate β₁ by methods of moments
    (_, Σ_Z) = SVAR_residuals(β₂_hat, Y, X)
    β₁_hat = estimate_β₁_MM(α, Σ_Z, icsvar)

    β_hat = vcat(β₁_hat, β₂_hat)

    if !compute_Σᵦ
        return β_hat, nothing
    end

    # 3 - Compute Σᵦ
    error("Not implemented.")
    #return (β_hat, Σᵦ_hat)
end

# Ainv(α, β₁) for trig parameterization
function get_Ainv(α::AbstractVector, β₁::AbstractVector{T}, icsvar::ICSVAR_triangular) where T
    return get_A(α, β₁, icsvar) \ I(icsvar.K)
end

# A(α,β₁) for trig parameterization
function get_A(α::AbstractVector{T}, β₁::AbstractVector{S}, icsvar::ICSVAR_triangular) where {T,S}
    k=0;
    L = Array{S}([ i>=j ? (k+=1; β₁[k]) : 0 for i=1:icsvar.K, j=1:icsvar.K ])
    k=0;
    U = Array{T}([ i<j ? (k+=1; α[k]) : 0 for i=1:icsvar.K, j=1:icsvar.K ])
    return L + U
end

# ∇A(α, β₁) for trig parameterization
function get_A_grad(α::AbstractVector{T}, β₁::AbstractVector{S}, l::Int64, icsvar::ICSVAR_triangular) where {T,S}
    if 1<=l & l<=length(α) # derivative w.r. to elements of α
        z = zeros(length(α))
        z[l] = 1.0
        k=0;
        U = Array{T}([ i<j ? (k+=1; z[k]) : 0 for i=1:icsvar.K, j=1:icsvar.K ])
        return U
    else # derivative w.r. to elements of β₁
        z = zeros(length(β₁))
        z[l-length(α)] = 1.0
        k=0;
        L = Array{S}([ i>=j ? (k+=1; z[k]) : 0 for i=1:icsvar.K, j=1:icsvar.K ])
        return L
    end
end


#
# Helper functions
#

function mmobj(σ::AbstractVector, α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_triangular)
    A = get_A(α, σ, icsvar)
    m = vech( I(icsvar.K) - A * Σ * A' ) # Alternative K*(K+1)/2 vector of moment conditions
    return m'*m 
end

# method-of-moments objective function for estimating β₁
function Triangular_MMobj(σ::AbstractVector, grad::AbstractVector, α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_triangular)
    if length(grad) > 0
        ForwardDiff.gradient!(grad, ξ -> mmobj(ξ, α, Σ, icsvar), σ)
    end
    return mmobj(σ, α, Σ, icsvar)
end

# estimation routine for β₁
function estimate_β₁_MM(α::AbstractVector, Σ::AbstractMatrix, icsvar::ICSVAR_triangular)
    
    # starting values
    σ_init = vech(I(icsvar.K) .+ 0.05)
    
    # Options
    opt = NLopt.Opt(:LD_LBFGS, icsvar.d_β₁)
    opt.min_objective = (σ, grad) -> Triangular_MMobj(σ, grad, α, Σ, icsvar)

    (f_hat, σ_hat, ret_val) = NLopt.optimize(opt, σ_init)
    if NLopt_success(ret_val)<1
        @printf("\nTriangular A: MM failed.", String(ret_val))
    end
    
    return σ_hat
end
