#---
# Pseudo Maximum Likelihood Estimators and Tests for Non-Gaussian SVARs
# Note: Tests & estimators assume a pseudo-density with fixed parameters for distribution of ϵ_k.
#---

using LinearAlgebra, StatsFuns, SpecialFunctions, ForwardDiff, NLopt

# Main interface for testing H0: α=α₀ vs. HA: α≠α₀
function PML_tests(α_H0::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, type::String, η::AbstractVector, init_vals::AbstractVector=[], bounds::AbstractVector=[[],[]], lm_only::Bool=false, verbose::Bool=false)

    # unrestricted PML estimator [H_A]
    nll_ur, θ_ur, _ = PML_estimator([], Y, X, icsvar, type, η, init_vals, bounds, verbose)

    # restricted PML estimator [H0: α = α_0]
    nll_r, θ_r, _ = PML_estimator(α_H0, Y, X, icsvar, type, η, init_vals, bounds, verbose)

    ## Compute Test Statistics
    if verbose
        @printf "PML Test Statistics\n"
    end

    R = vcat(I(icsvar.d_α), zeros(icsvar.d_β, icsvar.d_α))

    # (1) LM statistic
    g = ForwardDiff.jacobian(θ -> PML_loglik(θ[1:icsvar.d_α], θ[icsvar.d_α+1:end], Y, X, icsvar, η, type, true), θ_r) # score
    Σ = g' * g / icsvar.n
    HInv = ForwardDiff.hessian(θ -> PML_loglik(θ[1:icsvar.d_α], θ[icsvar.d_α+1:end], Y, X, icsvar, η, type, false), θ_r) \ I(icsvar.d_α+icsvar.d_β) # inv. hessian
    RCRInv = (R' * HInv * Σ * HInv * R) \ I(icsvar.d_α)
    V_θ_r = (R' * HInv * R) * RCRInv * (R' * HInv * R)'
    gbar = mean(g, dims=1)
    LM = icsvar.n * (R' * gbar') ⋅ (V_θ_r * (R' *gbar'))
    if verbose; @printf("... LM statistic\t%g\n", LM); end

    if lm_only; return LM; end
    
    # (2) Wald statistic
    g = ForwardDiff.jacobian(θ -> PML_loglik(θ[1:icsvar.d_α], θ[icsvar.d_α+1:end], Y, X, icsvar, η, type, true), θ_ur) # score
    HInv = ForwardDiff.hessian(θ -> PML_loglik(θ[1:icsvar.d_α], θ[icsvar.d_α+1:end], Y, X, icsvar, η, type, false), θ_ur) \ I(icsvar.d_α+icsvar.d_β) #inv. hessian
    Σ = g' * g / icsvar.n
    V_θ_ur = HInv * Σ * HInv
    δ = R' * θ_ur .- α_H0
    W = icsvar.n * δ' ⋅ ((R' * V_θ_ur * R) \ δ)
    if verbose; @printf("... Wald statistic\t%g\n", W); end

    # (3) LR statistics
    LR = -2 * icsvar.n * (nll_ur - nll_r) # eq. (12) of Lanne & Luoto (2021, JBES)
    if verbose; @printf("... LR statistic\t%g\n\n", LR); end

    return LM, W, LR
end


# Main interface for estimating γ=(α', β')'
function PML_estimator(α_H0::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, type::String, η::AbstractVector, init_vals::AbstractVector=[], bounds::AbstractVector=[[],[]], verbose::Bool=false)

    ## process inputs
    if verbose
        if !isempty(α_H0)
            @printf "Restricted PML Estimator\n"
        else
            @printf "Unrestricted PML Estimator\n"
        end
        @printf("... Distribution\t%s\n", type)
        @printf("... Dist. parameters\tη = %s\n", join([string(p) for p in η], ", "));
    end

    # set bounds for optimization procedure
    local θ_lb, θ_ub
    if length(bounds)!=2; error("Optimization bounds must be supplied as [[θ_lb], [θ_ub]]. Not supplied bounds will automatically be padded with -Inf/Inf."); end
    if length(bounds[1])==icsvar.d_α+icsvar.d_β && length(bounds[2])==icsvar.d_α+icsvar.d_β
        θ_lb = bounds[1]
        θ_ub = bounds[2]
    elseif length(bounds[1])==icsvar.d_α && length(bounds[2])==icsvar.d_α
        θ_lb = vcat(bounds[1], fill(-Inf, icsvar.d_β))
        θ_ub = vcat(bounds[2], fill(Inf, icsvar.d_β))
    else
        θ_lb = fill(-Inf, icsvar.d_α + icsvar.d_β)
        θ_ub = fill(Inf, icsvar.d_α + icsvar.d_β)
    end
    if verbose;
        @printf("... Bounds\t\tθ ∈ [[%s], [%s]]\n", join([string(p) for p in θ_lb], ", "), join([string(p) for p in θ_ub], ", "));
    end

    # set initial values
    local θ_init
    if isempty(init_vals);
        β_init, _ = estimate_β(α_H0, Y, X, icsvar, false) # OLS initial value
        θ_init = vcat(α_H0, β_init)
    else
        if length(init_vals)==icsvar.d_α
            β_init, _ = estimate_β(init_vals[1:icsvar.d_α], Y, X, icsvar, false) # OLS initial value
            θ_init = vcat(init_vals, β_init)
        elseif length(init_vals)==icsvar.d_α+icsvar.d_β
            θ_init = init_vals
        else
            error("Supplied initial values have wrong length.")
        end
    end
    if verbose;
        @printf("... Initialization\tθ = %s\n", join([string(p) for p in θ_init], ", "));
    end

    # check types
    if type!="t" && type!="gaussmix"
        error("Invalid distribution type: Pick 't' or 'gaussmix.")
    end

    ## PML Estimator
    local nll, θ_hat, ret_val
    if isempty(α_H0) # unrestricted PML estimator

        # options
        opt_ur = Opt(:LD_LBFGS, icsvar.d_α + icsvar.d_β)
        opt_ur.lower_bounds = θ_lb
        opt_ur.upper_bounds = θ_ub

        opt_ur.min_objective = (θ, grad) -> PML_obj_opt(θ, grad, Y, X, icsvar, η, type, false)
        (nll, θ_hat, ret_val) = NLopt.optimize(opt_ur, θ_init)
        if verbose & (NLopt_success(ret_val)<1)
            @printf("\nUnrestricted PML estimator optimization returned %s.", String(ret_val))
        end

    else # restricted PML estimator

        # options
        opt_r = Opt(:LD_LBFGS, icsvar.d_β)
        opt_r.lower_bounds = θ_lb[icsvar.d_α+1:end]
        opt_r.upper_bounds = θ_ub[icsvar.d_α+1:end]

        opt_r.min_objective = (β, grad) -> PML_obj_opt(vcat(α_H0, β), grad, Y, X, icsvar, η, type, true)
        (nll, β_hat, ret_val) = NLopt.optimize(opt_r, θ_init[icsvar.d_α+1:end])
        if verbose & (NLopt_success(ret_val)<1)
            @printf("\nRestricted PML estimator optimization returned %s.", String(ret_val))
        end
        θ_hat = vcat(α_H0, β_hat)

    end

    if verbose;
        @printf("... Result\t\t%s\n", String(ret_val));
        @printf("... Estimate\t\tθ = %s\n", join([string(p) for p in θ_hat], ", "));
        @printf("... Neg. log-lik.\tℓ = %g\n", nll);
        @printf("\n")
    end
    return nll, θ_hat, ret_val

end

# optimization objective function
function PML_obj_opt(θ::AbstractVector, grad::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, η::AbstractVector, type::String, H0::Bool)
    if length(grad) > 0
        if !H0
            ForwardDiff.gradient!(grad, ξ -> PML_loglik(ξ[1:icsvar.d_α], ξ[icsvar.d_α+1:end], Y, X , icsvar, η, type, false), θ)
        else
            ForwardDiff.gradient!(grad, β -> PML_loglik(θ[1:icsvar.d_α], β, Y, X , icsvar, η, type, false), θ[icsvar.d_α+1:end])
        end
    end
    return PML_loglik(θ[1:icsvar.d_α], θ[icsvar.d_α+1:end], Y, X, icsvar, η, type, false)
end

function PML_loglik(α::AbstractVector, β::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, η::AbstractVector, type::String, vector::Bool)
    β₁ = β[1:icsvar.d_β₁]
    β₂ = β[icsvar.d_β₁+1:end]

    Z, _ = SVAR_residuals(β₂, Y, X)
    A = get_A(α, β₁, icsvar)
    ϵ = A * Z

    # log-likelihood
    ll = similar(β₁, icsvar.n)
    ll[:] .= logabsdet(A)[1]
    @views @inbounds for i=1:icsvar.n, k=1:icsvar.K
        ll[i] += logpdf_std(ϵ[k, i], η[k], type)
    end

    if vector
        return ll
    else
        return -mean(ll)
    end
end


#
# Standardised densities
#

function logpdf_std(x::Real, η::AbstractVector, type::String)
    if type=="t"
        logpdf_std_t(x, 1/η[1])
    elseif type=="gaussmix"
        logpdf_std_2gm(x, η[1], η[2], η[3])
    else
        error(@sprintf("logpdf_std not implemented for type=%s.", type))
    end
end

# student-t, mean zero, unit variance (ν is 1/df)
function logpdf_std_t(x::Real, ν::Real)
    if ν == 0
        normlogpdf(x)
    else
        df = 1/ν # convert to degrees of freedom
        (log(df) - log(df-2))/2 + loggamma((df+1)/2) - loggamma(df/2)- log(π * df)/2 - (df+1)/2 * log(1 + x^2/(df-2))
    end
end
