#---
# Efficient scores and score test statistics
#---

using LinearAlgebra, StaticArrays, StatsFuns, SpecialFunctions, ForwardDiff

# Main interface
function ComputeScoreTest(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines, efficient::Bool)

    if efficient
        β_hat, _ = estimate_β_onestep(α, Y, X, icsvar, splines, false)
    else
        β_hat, _ = estimate_β(α, Y, X, icsvar, false)
    end

    Z_hat, _ = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)
    stat, rank = effscr_stat(α, β_hat[1:icsvar.d_β₁], β_hat[icsvar.d_β₁+1:icsvar.d_β], Z_hat, Y, X, icsvar, splines)
    
    return stat, rank
end

# Secondary interface receiving parameter estimates
function ComputeScoreTest(α::AbstractVector, β_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)

    Z_hat, _ = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)
    stat, rank = effscr_stat(α, β_hat[1:icsvar.d_β₁], β_hat[icsvar.d_β₁+1:icsvar.d_β], Z_hat, Y, X, icsvar, splines)
    
    return stat, rank
end

# one-step updated β estimates
function estimate_β_onestep(α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines, compute_Σᵦ::Bool=false)

    # 1 - obtain initial estimate from function provided by model declaration
    β_hat, _ = estimate_β(α, Y, X, icsvar, false)
    
    # 2 - efficient score update
    β_hat = effscore_update_β(β_hat, α, Y, X, icsvar, splines)
    if !compute_Σᵦ; return β_hat, nothing; end
    
    # 3 - compute Σᵦ
    Σᵦ_hat = effscore_compute_Σᵦ(β_hat, α, Y, X, icsvar, splines)
    return β_hat, Σᵦ_hat
end

# one-step efficient score updated estimates
function effscore_update_β(β_hat::AbstractVector, α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)

    Z_hat, _ = SVAR_residuals(β_hat[icsvar.d_β₁+1:icsvar.d_β], Y, X)

    # update parameters based on efficient score
    ℓ = build_effscr(α, β_hat[1:icsvar.d_β₁], β_hat[icsvar.d_β₁+1:icsvar.d_β], Z_hat, Y, X, icsvar, splines)
    I_ββ = ℓ[:, icsvar.d_α+1:end]' * ℓ[:, icsvar.d_α+1:end] / icsvar.n
    β_tilde = β_hat + I_ββ \ mean(ℓ[:, icsvar.d_α+1:end], dims = 1)'
    return vec(β_tilde)
end

# one-step efficient score updated estimates for γ
function effscore_update_γ(γ_hat::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)
    
    if length(γ_hat)!=icsvar.d_α+icsvar.d_β; error("γ has invalid length."); end
    
    # calculate efficient score
    Z_hat, _ = SVAR_residuals(γ_hat[icsvar.d_α+icsvar.d_β₁+1:icsvar.d_α+icsvar.d_β], Y, X)
    ℓ = build_effscr(γ_hat[1:icsvar.d_α], γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], γ_hat[icsvar.d_α+icsvar.d_β₁+1:icsvar.d_α+icsvar.d_β], Z_hat, Y, X, icsvar, splines)'
    
    # calculate efficient information matrix
    I_hat = ℓ * ℓ' / icsvar.n

    # updating step
    γ_updated = γ_hat + pinv(I_hat) * mean(ℓ, dims=2)

    return γ_updated
end

# one-step efficient score updated estimates
function effscore_compute_Σᵦ(β_tilde::AbstractVector, α::AbstractVector, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)

    # compute new residuals
    Z_tilde, _ = SVAR_residuals(β_tilde[icsvar.d_β₁+1:icsvar.d_β], Y, X)

    # compute covariance matrix from efficient info matrix
    ℓ_tilde = build_effscr(α, β_tilde[1:icsvar.d_β₁], β_tilde[icsvar.d_β₁+1:icsvar.d_β], Z_tilde, Y, X, icsvar, splines)
    Σᵦ_tilde = pinv( ℓ_tilde[:, icsvar.d_α+1:end]'*ℓ_tilde[:, icsvar.d_α+1:end])
    
    return Σᵦ_tilde
end


#
# 3 - Functions to compute efficient score statistic
#

# convenience interface that does not require pre-computation of efficient scores ℓ
# Input:  Z, Y, X have n cols
function effscr_stat(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, Z::AbstractMatrix, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)
    ℓ = build_effscr(α, β₁, β₂, Z, Y, X, icsvar, splines)
    (stat, infoRank) = compute_effscore_stat(length(α), ℓ)
    return (stat, infoRank)
end

function compute_effscore_stat(d_α::Int64, ℓ::AbstractMatrix)
    n = size(ℓ, 1)
    info = ℓ' * ℓ / n

    I_ββInv = pinv(info[d_α+1:end, d_α+1:end])

    g = I_ββInv * info[d_α+1:end, 1:d_α] # (I_ββ \ I_βα)
    κ_α = sum(ℓ[:, 1:d_α] - ℓ[:, d_α+1:end] * g, dims=1) # (ℓ_α - I_αβ * inv(I_ββ) * ℓ_β)'
    I_κ = (info[1:d_α, 1:d_α] - info[1:d_α, d_α+1:end] * g) # I_αα - I_αβ * (I_ββ \ I_βα)

    rank_I = rank(I_κ)
    I_κInv = pinv(I_κ)
    
    stat = (κ_α * I_κInv * κ_α') / n
    stat[1,1], rank_I
end


#
# 1 - Helper functions for computation of efficient scores
#

# Convenience interface, does not require pre-allocation of ℓ
# inputs: Z, Y, X have n rows
# output: ℓ remains n × d_α + d_β for compatibility with score test functions
function build_effscr(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, Z::AbstractMatrix{T}, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines) where T
    ℓ = zeros(T, size(Z, 2), length(α) + length(β₁) + length(β₂))
    build_effscr!(ℓ, α, β₁, β₂, permutedims(Z), permutedims(Y), permutedims(X), icsvar, splines)
    ℓ
end

# This takes a n × d_α + d_β matrix for ℓ
# The same orientation is needed for Z, Y, X; i.e. that they have n rows. 
function build_effscr!(ℓ::AbstractMatrix, α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, Z::AbstractMatrix, Y::AbstractMatrix, X::AbstractMatrix, icsvar::AbstractICSVAR, splines::BSplines)
    K = size(Y, 2)
    L = size(X, 2)
    d_α = length(α)
    d_β₁ = length(β₁)
    d_β₂ = length(β₂)

    A = get_A(α, β₁, icsvar)
    Ainv = get_Ainv(α, β₁, icsvar)
    ϵ = Z * A'
    κ = ϵ.^2 .- 1.0
    
    τ = broadcast(e -> get_τ(e, 1), eachcol(ϵ))
    ς = broadcast(e -> get_τ(e, 2), eachcol(ϵ))
    ζ = get_ζ(α, β₁, Ainv, icsvar)
    D = get_D(K, L)
    ϕs = ComputeDensityScores(ϵ, splines)
    μ = mean(X, dims = 1)
    Xtilde = X .- μ
    
    @inbounds @views for j=1:K
        for k=1:K
            if j==k
                τₖ = τ[k]
                for l = 1 : d_α + d_β₁
                    ℓ[:, l] .+= ζ[l][k, k] * (τₖ[1] * ϵ[:, k]  .+ τₖ[2] * κ[:, k])
                end
            else                
                for l = 1 : d_α + d_β₁   
                    ℓ[:, l] .+= ζ[l][k, j] * (ϕs[k] .* ϵ[:, j])
                end
                
            end
        end
        for l = 1:d_β₂
            ςⱼ = ς[j]
            AD = A[j, :]' *  D[l]
            XAD = Xtilde * -AD'
            ℓ[:, d_α + d_β₁+l] .+=  XAD .* ϕs[j]  .+ dot(AD, μ) * (ςⱼ[1] * ϵ[:, j] .+ ςⱼ[2] * κ[:, j])
        end
    end
end

function get_B_grad(K::Int64, L::Int64, l::Int64)
    dl = zeros(K*L)
    dl[l] = 1.0
    reshape(dl, K, L)
end

function get_ζ(α::AbstractVector, β₁::AbstractVector, Ainv::AbstractMatrix, icsvar::AbstractICSVAR)
    broadcast(l -> get_A_grad(α, β₁, l, icsvar) * Ainv, 1:length(α)+length(β₁))
end

function get_M(ϵₖ::AbstractVector)
    e3 = mean(ϵₖ.^3)
    SMatrix{2, 2}(1.0, e3, e3, mean(ϵₖ.^4) .- 1 )
end

function get_τ(ϵₖ::AbstractVector, l::Int64)
    if l == 1
        get_M(ϵₖ) \ SVector(0.0, -2.0)
    else
        get_M(ϵₖ) \ SVector(1.0, 0.0)
    end
end

function get_D(K::Int64, L::Int64)
    broadcast(l -> get_B_grad(K, L, l), 1:K*L)
end
