#---
# Functions to simulate ICSVARs for a given parameterization.
#---

using Distributions, LinearAlgebra, StatsBase, SparseArrays


# main function
function SimulateICSVAR(α::AbstractVector, β₁::AbstractVector, β₂::AbstractVector, ηs::Vector{String}, burnIn::Int64, icsvar::AbstractICSVAR, scaling::String)

    ηList = LoadDistributions()
    
    # 1 - sample independent structural errors
    nn = icsvar.n + icsvar.p + burnIn
    ϵ = Matrix{Float64}(undef, icsvar.K, nn)
    for k in 1:icsvar.K
        η_ix = min(k, length(ηs)) # use k-th or last entry in ηs list of distributions
        ϵ_k = rand(ηList[ηs[η_ix]], nn)
        if scaling=="sample"
            ϵ[k,:] = (ϵ_k' .- mean(ϵ_k)) / std(ϵ_k)
        elseif scaling=="population"
            ϵ[k,:] = (ϵ_k' .- mean(ηList[ηs[η_ix]])) / std(ηList[ηs[η_ix]])
        else
            error("Unknown 'scaling' option.")
        end
    end
    
    # 2 - sample SVAR
    Ainv = get_Ainv(α, β₁, icsvar)
    YY = zeros(icsvar.K, nn)
    sample_SVAR!(YY, ϵ, Ainv, β₂, icsvar.p, icsvar.constant)
    
    # 3 - return data
    Y = view(YY, :, burnIn+icsvar.p+1:nn)
    X = X_matrix(view(YY, :, burnIn+1:nn), icsvar.p, icsvar.constant)
    
    return Y, X
end


#
# Helper functions
#

# returns a list of pre-defined distributions
function LoadDistributions()
    return Dict(
        "N(0,1)" => Normal(),
        "t(15)" => TDist(15),
        "t(10)" => TDist(10),
        "t(5)" => TDist(5),
        "SKU" => MixtureModel(Normal, [(0, 1), (1/2, 4/9), (13/12, 25/81)], [1/5, 1/5, 3/5]),
        "KU" => MixtureModel(Normal, [(0, 1), (0, 1/100)], [2/3, 1/3]),
        "BM" => MixtureModel(Normal, [(-1, 4/9), (1, 4/9)], [1/2, 1/2]),
        "SPB" => MixtureModel(Normal, [(-2/3, 1/4), (2/3, 1/4)], [1/2, 1/2]),
        "SKB" => MixtureModel(Normal, [(0, 1), (2/3, 1/9)], [3/4, 1/4]),
        "TRI" => MixtureModel(Normal, [(-6/5, 9/25), (6/5, 9/25), (0, 1/16)], [9/20, 9/20, 1/10]),
    )
end

# Functions to sample SVAR given ϵ and Ainv; if initialised=true, assumes that Y[,1:p] are already set
function sample_zero_mean_SVAR!(Y::AbstractMatrix, ϵ::AbstractMatrix, Ainv::AbstractMatrix, Bs::AbstractVector, p::Int64, μ::AbstractVector, initialised::Bool)
    # simulate initial conditions
    if !initialised
        Y[:, 1] .= Ainv * ϵ[:, 1]
        # First p iterations
        @views @inbounds for i = 1:p-1       
            Y[:, i+1] = μ + Ainv * ϵ[:, i+1]
            for j=1:i
                Y[:, i+1] += Bs[j] * Y[:, 1+i-j]
            end
        end 
    end
    # simulate rest of the SVAR
    @views @inbounds for i = p+1:size(Y, 2)        
        Y[:, i] = μ + Ainv * ϵ[:, i]
        for j = 1:p
            Y[:, i] += Bs[j] * Y[:, i-j]
        end
    end 
end

# samples SVAR(p) given structural shock vector
function sample_SVAR!(Y::AbstractMatrix, ϵ::AbstractMatrix, Ainv::AbstractMatrix, β₂::AbstractVector, p::Int64, constant::Bool, initialised::Bool=false)
    K = size(Y, 1)
    if constant
        if p == 1
            B = SMatrix{K, K}(β₂[K+1:end])
            if !initialised
                μ = (I(K) - B) \  β₂[1:K]
            else
                μ = β₂[1:K]
            end
            sample_zero_mean_SVAR!(Y, ϵ, Ainv, [B], p, μ, initialised)
        else
            Bs = [SMatrix{K, K}(x) for x in Iterators.partition(β₂[K+1:end], K^2)]
            if !initialised
                μ = (I(K) - sum(Bs)) \ β₂[1:K]
            else
                μ = β₂[1:K]
            end
            sample_zero_mean_SVAR!(Y, ϵ, Ainv, Bs, p, μ, initialised)
        end
    else
        μ = zeros(K)
        if p == 1
            B = SMatrix{K, K}(β₂)
            sample_zero_mean_SVAR!(Y, ϵ, Ainv, [B], p, μ, initialised)
        else
            Bs = [SMatrix{K, K}(x) for x in Iterators.partition(β₂, K^2)]
            sample_zero_mean_SVAR!(Y, ϵ, Ainv, Bs, p, μ, initialised)
        end
    end
end

# Obtain bootstrap SVAR sample Y₁,...,Yₙ from structural residuals
# Note: e_boot, Y_boot are (K x n+p) matrices
function bootstrap_sample!(Y_boot::AbstractMatrix, ϵ_boot::AbstractMatrix, ϵ_hat::AbstractMatrix, Y::AbstractMatrix, Ainv::AbstractMatrix, β₂::AbstractVector, n::Int64, p::Int64, constant::Bool)
    K = size(ϵ_hat, 1)

    ϵ_boot[:,:] = ϵ_hat[sample(1:n, (K,n+p))] # re-sample the structural residuals (e_hat) separately to preserve independence  
    
    # sample block of initial conditions for the SVAR
    init = rand(1:(n-p+1))
    Y_init = Y[:, init:(init+p-1)]
    Y_boot[:,1:p] = Y_init
    
    # calculate implied SVAR observations based on simulated residuals and parameter values
    sample_SVAR!(Y_boot, ϵ_boot, Ainv, β₂, p, constant, true)
end
