#---
# SVAR(p) size simulation for competitor methods [Table 3]
#---

using CSV, DataFrames, Distributions, Statistics, Printf

include("../include/RunSimulation.jl")
include("../include/ICSVAR.jl") # load main functions
include("../include/Settings.jl") # load shared simulation settings
include("../include/PML.jl")
include("../include/DW.jl")

# simulation specifications
Base.@kwdef struct Job
    id::Int64
    η::String
    α::AbstractVector
    test::String
end

# simulation results for each specification
Base.@kwdef struct Result
    job::Job
    stats::DataFrame
end

# formats result to csv output
ResultHeader = "id,eta,alpha,statistic,rejection_rate"
function format_csv(result::Result)
    str = ""; i = 0;
    for row in CSV.RowWriter(result.stats)
        if i==0; i += 1; continue; end # skip header row
        str *= @sprintf("%d,\"%s\",%g,%s", result.job.id, result.job.η, result.job.α[1], row)
        i = i + 1
    end
    return str
end

# create jobs
function create_jobs()
    
    tests = ["score", "pml_t", "dw"]
    
    # local alternatives
    g_α = 21 # number of gridpoints
    αs =  LinRange(SIM_αs_caley[2][1] - .25, SIM_αs_caley[2][1] + .25, g_α)
    
    # create simulation workloads
    jobs = Job[]
    id = 1
    for η ∈ SIM_DISTS, α ∈ αs, test ∈ tests
        push!(jobs, Job( id=id, η=η, α=[α], test=test ))
        id += 1
    end

    return jobs
end

# simulation function
function run_job(job::Job)
    
    M = SIM_MC_REPS_SMALL

    # DGP settings
    n = 500
    K = 2
    p = 1
    R_type = "caley"
    α₀ = SIM_αs_caley[K]
    β₁=SIM_β₁s[K]
    β₂=SIM_β₂s[[K,p]]

    # test statistics
    testStats = Dict( 
        "score" => ["ols", "onestep"],
        "dw" => ["gmm", "bkr"],
        "pml_t" => ["lm"],
    )
    
    # skip alternative methods?
    skipAlternatives = false
    if "SKIP_ALTERNATIVE_TESTS" in keys(ENV)
        if parse(Int64, ENV["SKIP_ALTERNATIVE_TESTS"]) > 0
            skipAlternatives = true
        end
    end
    
    Random.seed!(SIM_JOB_SEED) # reset seed for each spec
    icsvar = ICSVAR_canonical( n=n, p=p, K=K, d_α=length(job.α), constant=SIM_CONSTANT, d_β₁=length(β₁), R_type=R_type )
    splines = BSplines(κ=4, B=SIM_SPLINES_B, adj=1.0, cutoff=SIM_SPLINES_CUTOFF)
    
    rejects = zeros(M, length(testStats[job.test]))
    for m in 1:M

        # simulate ICSVAR
        Y, X = SimulateICSVAR(job.α, β₁, β₂, [job.η], SIM_BURNIN, icsvar, "population")

        # 1 - Score Test
        if job.test=="score"
            # 1a - score test using OLS estimates
            stat_score_ols, rank_ols = ComputeScoreTest(α₀, Y, X, icsvar, splines, false)
            cv_ols = quantile(Chisq(rank_ols), 1-SIM_LEVEL)
            rejects[m, 1] = (stat_score_ols > cv_ols)
            
            # 1b - score test using one-step efficient estimates
            stat_score_onestep, rank_onestep = ComputeScoreTest(α₀, Y, X, icsvar, splines, true)
            cv_onestep = quantile(Chisq(rank_onestep), 1-SIM_LEVEL)
            rejects[m, 2] = (stat_score_onestep > cv_onestep)
        end

        # 2 - Drautzburg & Wright (2021) GMM & BKR tests
        if job.test=="dw"
            rejects[m,:] = [NaN, NaN]
            if !skipAlternatives
                stat_dw_gmm, stat_dw_bkr = ComputeDwTests(α₀, Y, X, icsvar)
                cv_dw_gmm, cv_dw_bkr = ComputeDwCritvals(α₀, Y, X, icsvar, SIM_BS_REPS, SIM_LEVEL)
                rejects[m, 1] = (stat_dw_gmm > cv_dw_gmm)
                rejects[m, 2] = (stat_dw_bkr > cv_dw_bkr)
            end
        end

        # 3 - Pseudo maximum likelihood tests: Fixed t distributions
        if job.test=="pml_t"
            rejects[m,:] = [NaN]
            if !skipAlternatives
                stat_pml_lm = PML_tests(α₀, Y, X, icsvar, "t", [[7], [12]], job.α, [[α₀[1]-.25], [α₀[1]+.25]], true, false)
                rejects[m, 1] = (stat_pml_lm > quantile(Chisq(icsvar.d_α), 1-SIM_LEVEL))
            end
        end
        
    end
    
    return Result(
        job = job,
        stats = DataFrame(
            testStatistic = [job.test * "_" * t for t in testStats[job.test]],
            rejectionRate = reshape( mean(rejects, dims=1), size(rejects,2) )
        )
    )
end

RunSimulation("Figure1.csv")