#---
# Application: Baumeister & Hamilton (2015)
# Main Parametrization
#---

using CSV, DataFrames, ProgressMeter, LaTeXStrings, Serialization

include("../include/ICSVAR.jl")
include("../include/ModelBH15.jl")
include("../include/IndependenceTests.jl")
include("../include/GMM.jl")
include("../include/Settings.jl")
Random.seed!(34575)


#
# 0 - Setup
#

Y, X, icsvar = load_model("direct")
splines = BSplines(B=4, adj=1.0, cutoff=0.05)
efficient = true
CS_levels =  [.33, .16, .05, .025]

# Grid: Structural elasticities α = (α^d, α^s)', imposing sign restrictions
n_grid = 10*EMP_GRID_FACTOR
α_ds = LinRange(-3, -1e-3, n_grid) # -
α_ss = LinRange(1e-3, 3, n_grid) # +
αs_grid = create_α_grid(α_ds, α_ss)
    

# 1 - Independence Test
function IndependenceTests()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid

    # (1) Estimate α via GMM using Keweloh moments, assuming strong identification
    α_init = [-.19, .76] # Initial values from Lanne & Luoto (2019)
    α_bounds = [[minimum(α_ds), minimum(α_ss)], [maximum(α_ds), maximum(α_ss)]]
    (_, θ_hat, _) = GMM_estimator([], Y, X, icsvar, 3, α_init, α_bounds) # 3 = Keweloh (2021, JBES) moments

    # (2) Save estimates of structural shocks at α_hat
    γ_hat = copy(θ_hat)
    ϵ_hat = get_A(γ_hat[1:icsvar.d_α], γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], icsvar) * SVAR_residuals(γ_hat[icsvar.d_α+icsvar.d_β₁+1:end], Y, X)[1]
    CSV.write(string(Base.@__DIR__)*"/../output/FigureS5.csv", DataFrame(ϵ_hat', ["eps1", "eps2"]))
    
    # test for independence of shocks using estimated α
    R = 10000
    pval_mopmq, pval_mt = IndependenceTests(γ_hat, Y, X, icsvar, R)
    @printf("Matteson & Tsay (2017): p-value = %g\n", round(pval_mt, digits=2))
    @printf("Montiel Olea, Plagborg-Moller and Qian (2022): p-value = %g\n\n", round(pval_mopmq, digits=2))
end


# 2 - Confidence Sets for elasticities from Algorithm 1
function CS_elasticities()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid
    
    # compute confidence set
    pb = Progress(length(αs_grid), "Figure 2 [1 proc]")
    CS_α_ind, _ = compute_CS_α(αs_grid, CS_levels, Y, X, icsvar, splines, efficient, pb) # Algorithm 1
    serialize(string(Base.@__DIR__)*"/../output/Figure2.dat", (CS_α_ind, αs_grid)) # save results
    
    # write confidence set to .csv
    (CS_α_ind, αs_grid) = deserialize(string(Base.@__DIR__)*"/../output/Figure2.dat")
    io = open(string(Base.@__DIR__)*"/../output/Figure2.csv", "w")
    write(io, @sprintf("alpha_d,alpha_s,level,covered\n"))
    for j in 1:length(CS_α_ind) # levels
        for α_ind in CS_α_ind[j]
           α_i = αs_grid[α_ind]
           write(io, @sprintf("%g,%g,%g,1\n", α_i[1], α_i[2], CS_levels[j]))
        end
    end
    close(io)
    
end


# 3 - Confidence Sets for IRFs from Algorithm 2
function CS_irfs()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid

    (CS_α_ind, αs_grid) = deserialize(string(Base.@__DIR__)*"/../output/Figure2.dat")

    # settings
    irfs = IrfSettings(
        K =             icsvar.K,
        H =             24, # max. horizon for IRFs
        γs =            CS_levels[[1,3]], # IRF CI confidence level
        cumulative =    trues(icsvar.K^2)
    )

    # compute IRF bands
    pb = Progress(sum(length.(CS_α_ind[[2,4]])), "Figure 3 [1 proc]")
    CS_irfs_lower, CS_irfs_upper = compute_IRF_bands(CS_α_ind[[2,4]], αs_grid, CS_levels[[2,4]], Y, X, icsvar, irfs, splines, efficient, pb) # Algorithm 2
    serialize(string(Base.@__DIR__)*"/../output/Figure3.dat", (CS_irfs_lower, CS_irfs_upper, icsvar, irfs));
    
    # write to .csv
    (CS_irfs_lower, CS_irfs_upper, icsvar, irfs) = deserialize(string(Base.@__DIR__)*"/../output/Figure3.dat");
    io = open(string(Base.@__DIR__)*"/../output/Figure3.csv", "w")
    write(io, @sprintf("level,variable,shock,h,lower,upper\n"))
    for l in 1:length(irfs.γs) # levels
        lower = reshape(minimum(CS_irfs_lower[l], dims=2), (icsvar.K^2, irfs.H+1))
        upper = reshape(maximum(CS_irfs_upper[l], dims=2), (icsvar.K^2, irfs.H+1))
        for j in 1:icsvar.K # shocks
            for i in 1:icsvar.K # vars
                for h in 0:irfs.H
                    write(io, @sprintf("%g,%d,%d,%d,%g,%g\n", CS_levels[[1,3]][l], i, j, h, lower[(j-1)*icsvar.K+i, h+1], upper[(j-1)*icsvar.K+i, h+1]))
                end
            end
        end
    end
    close(io)
    
end

@printf("Empirical Study - Baumeister & Hamilton (2015)\n\n")
IndependenceTests()
CS_elasticities()
CS_irfs()
