#---
# Application: Kilian and Murphy (2012)
#---

using CSV, DataFrames, ProgressMeter, LaTeXStrings, Serialization

include("../include/ICSVAR.jl")
include("../include/ModelKM12.jl")
include("../include/IndependenceTests.jl")
include("../include/GMM.jl")
include("../include/Settings.jl")
Random.seed!(34575)


#
# 0 - Setup
#

Y, X, icsvar = load_model()
splines = BSplines(B=4, adj=1.0, cutoff=0.05)
efficient = true
CS_levels =  [.33, .16, .05, .025]

# Grid: Structural parameters α = (α_qx, α_qp, α_xp)
n_α_qxs = 2*EMP_GRID_FACTOR
n_α_qps = 2*EMP_GRID_FACTOR
n_α_xps = EMP_GRID_FACTOR
α_qxs = LinRange(1e-6, .25, n_α_qxs) # + (short-run demand elasticity of supply)
α_qps = LinRange(1e-6, .1, n_α_qps) # + (short-run price elasticity of oil supply)
α_xps = LinRange(-3.0, -1e-6, n_α_xps) # -
αs_grid = create_α_grid(α_qxs, α_qps, α_xps)


# 1 - Independence Tests
function IndependenceTests()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid

    # (1) Estimate α via GMM using Keweloh moments, assuming strong identification
    α_init = [0.023/2.211, 0.128/5.526, -0.947] # Initial values from Kilian & Murphy (2012)
    β_init, _ = estimate_β(α_init, Y, X, icsvar, false)
    θ_bounds = [fill(-Inf, icsvar.d_α+icsvar.d_β₁+icsvar.d_β₂), fill(Inf, icsvar.d_α+icsvar.d_β₁+icsvar.d_β₂)]
    (_, θ_hat, _) = GMM_estimator([], Y, X, icsvar, 3, vcat(α_init, β_init), θ_bounds) # 3 = Keweloh (2021, JBES) moments
    
    # (2) Save estimates of structural shocks at α_hat
    γ_hat = copy(θ_hat)
    ϵ_hat = get_A(γ_hat[1:icsvar.d_α], γ_hat[icsvar.d_α+1:icsvar.d_α+icsvar.d_β₁], icsvar) * SVAR_residuals(γ_hat[icsvar.d_α+icsvar.d_β₁+1:end], Y, X)[1]
    CSV.write(string(Base.@__DIR__)*"/../output/FigureS6.csv", DataFrame(ϵ_hat', ["eps1", "eps2", "eps3"]))

    # test for independence of shocks using estimated α
    R = 10000
    pval_mopmq, pval_mt = IndependenceTests(γ_hat, Y, X, icsvar, R)
    @printf("Matteson & Tsay (2017): p-value = %g\n", round(pval_mt, digits=2))
    @printf("Montiel Olea, Plagborg-Moller and Qian (2022): p-value = %g\n\n", round(pval_mopmq, digits=2))
end


# 2 - Confidence Sets for elasticities from Algorithm 1
function CS_elasticities()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid

    # compute confidence set
    pb = Progress(length(αs_grid), "Figure 4 [1 proc]")
    CS_α_ind, _ = compute_CS_α(αs_grid, CS_levels, Y, X, icsvar, splines, efficient, pb) # Algorithm 1
    serialize(string(Base.@__DIR__)*"/../output/Figure4.dat", (CS_α_ind, αs_grid)) # save results

    # write confidence set to .csv
    (CS_α_ind, αs_grid) = deserialize(string(Base.@__DIR__)*"/../output/Figure4.dat")
    io = open(string(Base.@__DIR__)*"/../output/Figure4.csv", "w")
    write(io, @sprintf("alpha_qx,alpha_qp,alpha_xp,level,covered\n"))
    for j in 1:length(CS_α_ind) # levels
        for α_ind in CS_α_ind[j]
           α_i = αs_grid[α_ind]
           write(io, @sprintf("%g,%g,%g,%g,1\n", α_i[1], α_i[2], α_i[2], CS_levels[j]))
        end
    end
    close(io)
    
end


# 3 - Confidence Sets for IRFs 
function CS_irfs()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid
    
    (CS_α_ind, αs_grid) = deserialize(string(Base.@__DIR__)*"/../output/Figure4.dat")

    irfs = IrfSettings(
        K =             icsvar.K,
        H =             16, # max. horizon for IRFs
        γs =            CS_levels[[1,3]], # IRF CI confidence level
        cumulative =    (mod.(1:icsvar.K^2,3) .== 1) # responses of Δq
    )

    # compute IRF bands
    pb = Progress(sum(length.(CS_α_ind[[2,4]])), "Figure 5 [1 proc]")
    CS_irfs_lower, CS_irfs_upper = compute_IRF_bands(CS_α_ind[[2,4]], αs_grid, CS_levels[[2,4]], Y, X, icsvar, irfs, splines, efficient, pb) # Algorithm 2
    serialize(string(Base.@__DIR__)*"/../output/Figure5.dat", (CS_irfs_lower, CS_irfs_upper, icsvar, irfs));
    
    # write to .csv
    (CS_irfs_lower, CS_irfs_upper, icsvar, irfs) = deserialize(string(Base.@__DIR__)*"/../output/Figure5.dat");
    io = open(string(Base.@__DIR__)*"/../output/Figure5.csv", "w")
    write(io, @sprintf("level,variable,shock,h,lower,upper\n"))
    for l in 1:length(irfs.γs) # levels
        lower = reshape(minimum(CS_irfs_lower[l], dims=2), (icsvar.K^2, irfs.H+1))
        upper = reshape(maximum(CS_irfs_upper[l], dims=2), (icsvar.K^2, irfs.H+1))
        for j in 1:icsvar.K # shocks
            for i in 1:icsvar.K # vars
                for h in 0:irfs.H
                    write(io, @sprintf("%g,%d,%d,%d,%g,%g\n", CS_levels[[1,3]][l], i, j, h, lower[(j-1)*icsvar.K+i, h+1], upper[(j-1)*icsvar.K+i, h+1]))
                end
            end
        end
    end
    close(io)

end

@printf("Empirical Study - Kilian & Murphy (2012)\n\n")
IndependenceTests()
CS_elasticities()
CS_irfs()
