#---
# SVAR(p) coverage & average length simulation [Figures S1 & S2]
#---

using CSV, DataFrames, Distributions, Statistics, Printf

include("../include/RunSimulation.jl")
include("../include/ICSVAR.jl") # load main functions
include("../include/Settings.jl") # load shared simulation settings
include("../include/DW.jl")

# simulation specifications
Base.@kwdef struct Job
    id::Int64
    n::Int64
    p::Int64
    K::Int64
    H::Int64
    η::String
end

# simulation results for each specification
Base.@kwdef struct Result
    job::Job
    coverage_α_score_ols::Real
    coverage_α_score_onestep::Real
    coverage_α_dw_gmm::Real
    coverage_α_dw_bkr::Real
    irfs::DataFrame
end

# formats result to csv output
ResultHeader = "id,n,p,K,eta,h,i,j,coverage_score_ols,length_score_ols,coverage_score_onestep,length_score_onestep,coverage_dw_gmm,length_dw_gmm,coverage_dw_bkr,length_dw_bkr"
function format_csv(result::Result)
    # coverage rates for α
    str = @sprintf("%d,%d,%d,%d,\"%s\",%d,,,%g,,%g,,%g,,%g,\n", result.job.id, result.job.n, result.job.p, result.job.K, result.job.η, -1, result.coverage_α_score_ols, result.coverage_α_score_onestep, result.coverage_α_dw_gmm, result.coverage_α_dw_bkr)
    i = 0
    for row in CSV.RowWriter(result.irfs)
        if i==0; i += 1; continue; end # skip header row
        str *= @sprintf("%d,%d,%d,%d,\"%s\",%s", result.job.id, result.job.n, result.job.p, result.job.K, result.job.η, row)
        i = i + 1
    end
    return str
end

# create jobs
function create_jobs()

    # create simulation workloads
    jobs = Job[]
    id = 1
    n = 500
    p = 1
    K = 2
    H = 12
    for η ∈ SIM_DISTS
        push!(jobs, Job( id=id, n=n, p=p, K=K, H=H, η=η ))
        id += 1
    end

    return jobs
end

# simulation function
function run_job(job::Job)
    
    M = SIM_MC_REPS_SMALL
    
    # DGP settings
    R_type = "caley"
    α₀ = SIM_αs_caley[job.K]
    β₁ = SIM_β₁s[job.K]
    β₂ = SIM_β₂s[[job.K,job.p]]
    γ₁s = [.05]
    γ₂s = [.05]
    αs_grid = create_α_grid(range(α₀[1]-.25, α₀[1]+.25, length=21))

    irfs = IrfSettings(
        K =             job.K,
        H =             job.H, # max. horizon for IRFs
        γs =            [.1], # IRF CI confidence level
        γ₁s =           γ₁s,
        γ₂s =           γ₂s,
        cumulative =    trues(job.K^2)
    )

    # skip alternative methods?
    skipAlternatives = false
    if "SKIP_ALTERNATIVE_TESTS" in keys(ENV)
        if parse(Int64, ENV["SKIP_ALTERNATIVE_TESTS"]) > 0
            skipAlternatives = true
        end
    end
    
    Random.seed!(SIM_JOB_SEED) # reset seed for each spec
    icsvar = ICSVAR_canonical( n=job.n, p=job.p, K=job.K, d_α=length(α₀), constant=SIM_CONSTANT, d_β₁=length(β₁), R_type=R_type )
    splines = BSplines(κ=4, B=SIM_SPLINES_B, adj=1.0, cutoff=SIM_SPLINES_CUTOFF)

    # compute true irf for coverage calculations
    irf_true = compute_irfs(α₀, β₁, β₂, icsvar, irfs)
    
    coverage_α = fill(NaN, (M, 4))
    coverage_irfs = fill(NaN, (M, job.K^2*(job.H+1), 4))
    length_irfs = fill(NaN, (M, job.K^2*(job.H+1), 4))
    for m in 1:M

        # simulate ICSVAR
        Y, X = SimulateICSVAR(α₀, β₁, β₂, [job.η], SIM_BURNIN, icsvar, "population")
        
        #
        # 1 = Construct IRF CS from score tests
        #
        
        # 1a - Using OLS estimates
        CS_α_ind, _ = compute_CS_α(αs_grid, γ₁s, Y, X, icsvar, splines, false) # Algorithm 1
        if !isempty(CS_α_ind[1])
            CS_irfs_lower, CS_irfs_upper = compute_IRF_bands(CS_α_ind, αs_grid, γ₂s, Y, X, icsvar, irfs, splines, false) # Algorithm 2
            coverage_α[m,1] = 11 in CS_α_ind[1]
            coverage_irfs[m,:,1] = check_coverage(CS_irfs_lower[1], CS_irfs_upper[1], irf_true, irfs.K, irfs.H)
            length_irfs[m,:,1] = vec( maximum(CS_irfs_upper[1], dims=2) - minimum(CS_irfs_lower[1], dims=2) )
        end

        # 1b - Using one-step efficient estimates
        CS_α_ind, _ = compute_CS_α(αs_grid, γ₁s, Y, X, icsvar, splines, true) # Algorithm 1
        if !isempty(CS_α_ind[1])
            CS_irfs_lower, CS_irfs_upper = compute_IRF_bands(CS_α_ind, αs_grid, γ₂s, Y, X, icsvar, irfs, splines, true) # Algorithm 2
            coverage_α[m,2] = 11 in CS_α_ind[1]
            coverage_irfs[m,:,2] = check_coverage(CS_irfs_lower[1], CS_irfs_upper[1], irf_true, irfs.K, irfs.H)
            length_irfs[m,:,2] = vec( maximum(CS_irfs_upper[1], dims=2) - minimum(CS_irfs_lower[1], dims=2) )
        end
        
        # GMM and BKR tests
        if !skipAlternatives
            CS_α_ind, _ = compute_CS_α_DW(αs_grid, γ₁s, Y, X, icsvar, 500)
            if !isempty(CS_α_ind[1]) # GMM
                CS_irfs_lower_gmm, CS_irfs_upper_gmm = compute_IRF_bands([CS_α_ind[1]], αs_grid, γ₂s, Y, X, icsvar, irfs, splines, false) # Algorithm 2
                coverage_α[m,3] = 11 in CS_α_ind[1]
                coverage_irfs[m,:,3] = check_coverage(CS_irfs_lower_gmm[1], CS_irfs_upper_gmm[1], irf_true, irfs.K, irfs.H)
                length_irfs[m,:,3] = vec( maximum(CS_irfs_upper_gmm[1], dims=2) - minimum(CS_irfs_lower_gmm[1], dims=2) )
            end
            if !isempty(CS_α_ind[2]) # BKR
                CS_irfs_lower_bkr, CS_irfs_upper_bkr = compute_IRF_bands([CS_α_ind[2]], αs_grid, γ₂s, Y, X, icsvar, irfs, splines, false) # Algorithm 2
                coverage_α[m,4] = 11 in CS_α_ind[2]
                coverage_irfs[m,:,4] = check_coverage(CS_irfs_lower_bkr[1], CS_irfs_upper_bkr[1], irf_true, irfs.K, irfs.H)
                length_irfs[m,:,4] = vec( maximum(CS_irfs_upper_bkr[1], dims=2) - minimum(CS_irfs_lower_bkr[1], dims=2) )
            end
        end

    end

    return Result(
        job = job,
        coverage_α_score_ols = mapslices(nanmean, coverage_α[:,1], dims=1)[1],
        coverage_α_score_onestep = mapslices(nanmean, coverage_α[:,2], dims=1)[1],
        coverage_α_dw_gmm = mapslices(nanmean, coverage_α[:,3], dims=1)[1],
        coverage_α_dw_bkr = mapslices(nanmean, coverage_α[:,4], dims=1)[1],
        irfs = DataFrame(
            H=repeat(0:irfs.H, inner=job.K^2),
            i=repeat(repeat(1:job.K,outer=job.K),outer=irfs.H+1),
            j=repeat(repeat(1:job.K,inner=job.K),outer=irfs.H+1),
            coverage_score_ols = reshape( mapslices(nanmean, coverage_irfs[:,:,1], dims=1), size(coverage_irfs,2) ),
            length_score_ols = reshape( mapslices(nanmean, length_irfs[:,:,1], dims=1), size(length_irfs,2) ),
            coverage_score_onestep = reshape( mapslices(nanmean, coverage_irfs[:,:,2], dims=1), size(coverage_irfs,2) ),
            length_score_onestep = reshape( mapslices(nanmean, length_irfs[:,:,2], dims=1), size(length_irfs,2) ),
            coverage_dw_gmm = reshape( mapslices(nanmean, coverage_irfs[:,:,3], dims=1), size(coverage_irfs,2) ),
            length_dw_gmm = reshape( mapslices(nanmean, length_irfs[:,:,3], dims=1), size(length_irfs,2) ),
            coverage_dw_bkr = reshape( mapslices(nanmean, coverage_irfs[:,:,4], dims=1), size(coverage_irfs,2) ),
            length_dw_bkr = reshape( mapslices(nanmean, length_irfs[:,:,4], dims=1), size(length_irfs,2) )
        )
    )
end

RunSimulation("FigureS1_S2.csv")