#---
# Application: Baumeister & Hamilton (2015)
# Alternative Parametrization
#---

using CSV, DataFrames, ProgressMeter, LaTeXStrings, Serialization

include("../include/ICSVAR.jl")
include("../include/ModelBH15.jl")
include("../include/Settings.jl")
Random.seed!(34575)


#
# 0 - Setup
#

Y, X, icsvar = load_model("euler")
splines = BSplines(B=4, adj=1.0, cutoff=0.05)
efficient = true
CS_levels =  [.33, .05]

# Grid 1: Rotation angle α, imposing sign restrictions
n_grid = 10*EMP_GRID_FACTOR
αs_grid = create_α_grid( LinRange(1e-3, pi/2, n_grid) )

# Grid 2: Structural elasticities g(α, β) = (α^d, α^s)', imposing sign restrictions
n_grid = 10*EMP_GRID_FACTOR
g_ds = LinRange(-3, -1e-3, n_grid) # -
g_ss = LinRange(1e-3, 3, n_grid) # +
gs_grid = create_α_grid(g_ds, g_ss)

# 1 - Confidence Sets for elasticities from Algorithms 1 & 2
function CS_elasticities()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid, gs_grid

    # Algorithm 1 - CS for rotation angle α
    pb1 = Progress(length(αs_grid), "Figure S3 (1/2) [1 proc]")
    CS_α_ind, _ = compute_CS_α(αs_grid, CS_levels./2, Y, X, icsvar, splines, efficient, pb1)
    serialize(string(Base.@__DIR__)*"/../output/FigureS3.dat", (CS_α_ind, αs_grid)) # save results
    
    # Algorithm 2 - CS for structural elasticities
    pb2 = Progress(sum(length.(CS_α_ind)), "Figure S3 (2/2) [1 proc]")
    CS_g_ind = compute_CS_elasticities(CS_α_ind, αs_grid, gs_grid, CS_levels./2, Y, X, icsvar, splines, efficient, pb2)

    # write confidence set to .csv
    io = open(string(Base.@__DIR__)*"/../output/FigureS3.csv", "w")
    write(io, @sprintf("alpha_d,alpha_s,level,covered\n"))
    for j in 1:length(CS_g_ind) # levels
        for g_ind in CS_g_ind[j]
            g_i = gs_grid[g_ind]
            write(io, @sprintf("%g,%g,%g,1\n", g_i[1], g_i[2], CS_levels[j]))
        end
    end
    close(io)
    
end

# 2 - Confidence Sets for IRFs from Algorithm 2
function CS_irfs()

    global Y, X, icsvar, splines, efficient, CS_levels, αs_grid

    (CS_α_ind, αs_grid) = deserialize(string(Base.@__DIR__)*"/../output/FigureS3.dat") # save results

    # settings
    irfs = IrfSettings(
        K =             icsvar.K,
        H =             24, # max. horizon for IRFs
        γs =            CS_levels, # IRF CI confidence level
        cumulative =    trues(icsvar.K^2)
    )

    # compute IRF bands
    pb = Progress(sum(length.(CS_α_ind)), "Figure S4 [1 proc]")
    CS_irfs_lower, CS_irfs_upper = compute_IRF_bands(CS_α_ind, αs_grid, CS_levels./2, Y, X, icsvar, irfs, splines, efficient, pb) # Algorithm 2
    serialize(string(Base.@__DIR__)*"/../output/FigureS4.dat", (CS_irfs_lower, CS_irfs_upper, icsvar, irfs));

    # write to .csv
    (CS_irfs_lower, CS_irfs_upper, icsvar, irfs) = deserialize(string(Base.@__DIR__)*"/../output/FigureS4.dat");
    
    io = open(string(Base.@__DIR__)*"/../output/FigureS4.csv", "w")
    write(io, @sprintf("level,variable,shock,h,lower,upper\n"))
    for l in 1:length(irfs.γs) # levels
        lower = reshape(minimum(CS_irfs_lower[l], dims=2), (icsvar.K^2, irfs.H+1))
        upper = reshape(maximum(CS_irfs_upper[l], dims=2), (icsvar.K^2, irfs.H+1))
        for j in 1:icsvar.K # shocks
            for i in 1:icsvar.K # vars
                for h in 0:irfs.H
                    write(io, @sprintf("%g,%d,%d,%d,%g,%g\n", CS_levels[l], i, j, h, lower[(j-1)*icsvar.K+i, h+1], upper[(j-1)*icsvar.K+i, h+1]))
                end
            end
        end
    end
    close(io)

end

@printf("Empirical Study - Baumeister & Hamilton (2015) Alternative Parametrization\n\n")
CS_elasticities()
CS_irfs()
