#---
# Table S1: Empirical rejection frequencies: Alternative parametrization for A(α,σ).
#---

using Distributions, Statistics, Printf

include("../include/RunSimulation.jl")
include("../include/ICSVAR.jl") # load main functions
include("../include/Settings.jl") # load shared simulation settings

# simulation specifications
Base.@kwdef struct Job
    id::Int64
    n::Int64
    p::Int64
    K::Int64
    η::String
    α::AbstractVector
    β₁::AbstractVector
    β₂::AbstractVector
end

# simulation results for each specification
Base.@kwdef struct Result
    job::Job
    rejectionRateOLS::Real
    rejectionRateOneStep::Real
end

# formats result to csv output
ResultHeader = "id,n,p,K,eta,score_ols,score_onestep"
function format_csv(result::Result)
    return @sprintf("%d,%d,%d,%d,\"%s\",%g,%g\n", result.job.id, result.job.n, result.job.p, result.job.K, result.job.η, result.rejectionRateOLS, result.rejectionRateOneStep)
end

# create jobs
function create_jobs()

    # specifications to run
    ns = [200, 500, 1000]
    ps = [1, 4, 12]
    Ks = [2, 3]
    
    # create simulation workloads
    jobs = Job[]
    id = 1
    for n ∈ ns, p ∈ ps, η ∈ SIM_DISTS, K ∈ Ks
        push!(jobs, Job( id=id, n=n, p=p, K=K, η=η, α=SIM_αs_caley[K], β₁=SIM_β₁s[K], β₂=SIM_β₂s[[K,p]] ))
        id += 1
    end

    return jobs
end

# simulation function
function run_job(job::Job)
    
    M = SIM_MC_REPS

    Random.seed!(SIM_JOB_SEED) # reset seed for each spec
    icsvar = ICSVAR_triangular( n=job.n, p=job.p, K=job.K, d_α=length(job.α), constant=SIM_CONSTANT, d_β₁=length(job.β₁) )
    splines = BSplines(κ=4, B=SIM_SPLINES_B, adj=1.0, cutoff=SIM_SPLINES_CUTOFF)
    
    rejects = falses(M, 2) # [ols, one-step]
    for m in 1:M

        # 1 - simulate ICSVAR
        Y, X = SimulateICSVAR(job.α, job.β₁, job.β₂, [job.η], SIM_BURNIN, icsvar, "population")

        # 2 - score test using OLS estimates
        stat_score_ols, rank_ols = ComputeScoreTest(job.α, Y, X, icsvar, splines, false)
        cv_ols = quantile(Chisq(rank_ols), 1-SIM_LEVEL)
        rejects[m, 1] = (stat_score_ols > cv_ols)
        
        # 3 - score test using one-step efficient estimates
        stat_score_onestep, rank_onestep = ComputeScoreTest(job.α, Y, X, icsvar, splines, true)
        cv_onestep = quantile(Chisq(rank_onestep), 1-SIM_LEVEL)
        rejects[m, 2] = (stat_score_onestep > cv_onestep)
        
    end
    rejectionRates = mean(rejects, dims=1)

    return Result( job=job, rejectionRateOLS=rejectionRates[1], rejectionRateOneStep=rejectionRates[2] )
end

RunSimulation("TableS1.csv")