#---
# Table S4: Empirical rejection frequencies for larger SVARs
#---

using Distributions, Statistics, Printf

include("../include/RunSimulation.jl")
include("../include/ICSVAR.jl") # load main functions
include("../include/Settings.jl") # load shared simulation settings

# simulation specifications
Base.@kwdef struct Job
    id::Int64
    n::Int64
    p::Int64
    K::Int64
    η::String
    α::AbstractVector
    β₁::AbstractVector
    β₂::AbstractVector
end

# simulation results for each specification
Base.@kwdef struct Result
    job::Job
    rejectionRate::Real
end

# formats result to csv output
ResultHeader = "id,n,p,K,eta,score"
function format_csv(result::Result)
    return @sprintf("%d,%d,%d,%d,\"%s\",%g\n", result.job.id, result.job.n, result.job.p, result.job.K, result.job.η, result.rejectionRate)
end

# create jobs
function create_jobs()

    # specifications to run
    ns = [200, 500, 1000]
    ps = [1, 4, 12]
    Ks = [2, 3]
    
    # create simulation workloads
    jobs = Job[]
    id = 1
    for n ∈ ns, p ∈ ps, η ∈ SIM_DISTS, K ∈ Ks
        push!(jobs, Job( id=id, n=n, p=p, K=K, η=η, α=SIM_αs_caley[K], β₁=SIM_β₁s[K], β₂=SIM_β₂s[[K,p]] ))
        id += 1
    end

    return jobs
end

# simulation function
function run_job(job::Job)
    
    M = SIM_MC_REPS
    maxTries = 5
    
    Random.seed!(SIM_JOB_SEED) # reset seed for each spec
    icsvar = ICSVAR_canonical( n=job.n, p=job.p, K=job.K, d_α=length(job.α), constant=SIM_CONSTANT, d_β₁=length(job.β₁), R_type="caley" )
    splines = BSplines(κ=4, B=51, adj=1.0, cutoff=SIM_SPLINES_CUTOFF, crossValidate=true)
    
    rejects = falses(M)
    for m in 1:M

        for l in 1:maxTries
            try
                # 1 - simulate ICSVAR
                Y, X = SimulateICSVAR(job.α, job.β₁, job.β₂, [job.η], SIM_BURNIN, icsvar, "population")
                
                # 2 - compute score test
                stat_score, rank_score = ComputeScoreTest(job.α, Y, X, icsvar, splines, false)
                cv_score = 0.0
                if rank_score>0; cv_score = quantile(Chisq(rank_score), 1-SIM_LEVEL); end
                rejects[m] = (stat_score > cv_score)
                break
            catch e
                @sprintf("Restarting MC iteration %d", m)
            end
        end
        
    end
    rejectionRates = mean(rejects)

    return Result( job=job, rejectionRate=rejectionRates )
end

RunSimulation("TableS3.csv")