#---
# Table S5 - Efficiency of one-step updates
#---

using CSV, DataFrames, Distributions, Statistics, Printf

include("../include/RunSimulation.jl")
include("../include/ICSVAR.jl") # load main functions
include("../include/Settings.jl") # load shared simulation settings
include("../include/PML.jl")
include("../include/GMM.jl")

# simulation specifications
Base.@kwdef struct Job
    id::Int64
    η::String
    method::String
end

# simulation results for each specification
Base.@kwdef struct Result
    job::Job
    performance::DataFrame
end

# formats result to csv output
ResultHeader = "id,eta,method,type,rmse"
function format_csv(result::Result)
    str = ""; i = 0;
    for row in CSV.RowWriter(result.performance)
        if i==0; i += 1; continue; end # skip header row
        str *= @sprintf("%d,\"%s\",%s", result.job.id, result.job.η, row)
        i = i + 1
    end
    return str
end

# create jobs
function create_jobs()

    # methods
    methods = ["pml_t", "gmm_ll", "gmm_kew"]
    
    # create simulation workloads
    jobs = Job[]
    id = 1
    for η ∈ SIM_DISTS, method ∈ methods
        push!(jobs, Job( id=id, η=η, method=method ))
        id += 1
    end

    return jobs
end

# simulation function
function run_job(job::Job)

    M = SIM_MC_REPS

    # DGP settings
    R_type = "caley"
    n = 500
    K = 2
    p = 1
    α₀ = SIM_αs_caley[K]
    β₁ = SIM_β₁s[K]
    β₂ = SIM_β₂s[[K,p]]

    Random.seed!(SIM_JOB_SEED) # reset seed for each spec
    icsvar = ICSVAR_canonical( n=n, p=p, K=K, d_α=length(α₀), constant=SIM_CONSTANT, d_β₁=length(β₁), R_type=R_type )
    splines = BSplines(κ=4, B=SIM_SPLINES_B, adj=1.0, cutoff=SIM_SPLINES_CUTOFF)
    
    sqerrors = zeros(M, 2) # original & one-step update
    for m in 1:M

        # simulate ICSVAR
        Y, X = SimulateICSVAR(α₀, β₁, β₂, [job.η], SIM_BURNIN, icsvar, "population")

        # use starting value based on true α
        α_init = copy(α₀) .+ .05
        local γ_hat

        # 2a - PML estimator of Gourieroux, Monfort, Renne (2021) using t(7) and t(12) distributions
        if job.method=="pml_t"
            _, γ_hat, _ = PML_estimator([], Y, X, icsvar, "t", [[7], [12]], α_init, [[α₀[1]-.25], [α₀[1]+.25]], false)
        end
        
        # 2b - GMM estimator of Lanne & Luoto (2021) using one cu-kurtosis condition
        if job.method=="gmm_ll"
            _, γ_hat, _ = GMM_estimator([], Y, X, icsvar, 1, α_init, [[α₀[1]-.25], [α₀[1]+.25]], false)
        end

        # 2c - GMM estimator of Keweloh (2021) using both cu-kurtosis condition
        if job.method=="gmm_kew"
            _, γ_hat, _ = GMM_estimator([], Y, X, icsvar, 3, α_init, [[α₀[1]-.25], [α₀[1]+.25]], false)
        end
        
        # 3 - One step update
        γ_osu = effscore_update_γ(γ_hat, Y, X, icsvar, splines)

        # 4 - Save performance
        sqerrors[m,1] = (α₀[1] - γ_hat[1])^2
        sqerrors[m,2] = (α₀[1] - γ_osu[1])^2
        
    end

    return Result(
        job = job,
        performance = DataFrame(
            method = [job.method for j in 1:2],
            type = ["original", "onestep_update"],
            rmse = reshape(sqrt.(mean(sqerrors, dims=1)), 2),
        )
    )
end

RunSimulation("TableS5.csv")